import datetime
import pandas as pd
import pytest

from thermostat_nw.importers import (
    from_csv,
    normalize_utc_offset,
)

from thermostat_nw.util.testing import get_data_path

from .fixtures.single_stage import (
    thermostat_type_1,
    thermostat_type_1_utc,
    thermostat_type_1_utc_bad,
    thermostat_type_1_too_many_minutes,
    thermostat_type_1_zip_bad,
    thermostat_type_1_data_out_of_order,
    thermostat_type_1_data_missing_header,
    thermostat_type_1_cache,
)


def test_import_csv(thermostat_type_1):
    def assert_is_series_with_shape(series, shape):
        assert isinstance(series, pd.Series)
        assert series.shape == shape

    assert_is_series_with_shape(thermostat_type_1.cool_runtime_daily, (1461,))
    assert_is_series_with_shape(thermostat_type_1.heat_runtime_daily, (1461,))

    assert_is_series_with_shape(thermostat_type_1.cool_runtime_hourly, (35064,))
    assert_is_series_with_shape(thermostat_type_1.heat_runtime_hourly, (35064,))

    assert_is_series_with_shape(thermostat_type_1.auxiliary_heat_runtime, (35064,))
    assert_is_series_with_shape(thermostat_type_1.emergency_heat_runtime, (35064,))

    assert_is_series_with_shape(thermostat_type_1.temperature_in, (35064,))
    assert_is_series_with_shape(thermostat_type_1.temperature_out, (35064,))


def test_import_csv_cache(thermostat_type_1_cache):
    assert thermostat_type_1_cache is not None


def test_utc_offset(thermostat_type_1_utc, thermostat_type_1_utc_bad):
    assert normalize_utc_offset("+0") == datetime.timedelta(0)
    assert normalize_utc_offset("-0") == datetime.timedelta(0)
    assert normalize_utc_offset("0") == datetime.timedelta(0)
    assert normalize_utc_offset(0) == datetime.timedelta(0)
    assert normalize_utc_offset("+6") == datetime.timedelta(0, 21600)
    assert normalize_utc_offset("-6") == datetime.timedelta(-1, 64800)
    assert normalize_utc_offset(-6) == datetime.timedelta(-1, 64800)

    with pytest.raises(TypeError) as excinfo:
        normalize_utc_offset(6)
    assert "Invalid UTC" in str(excinfo)

    # Load a thermostat with utc offset == 0
    assert isinstance(thermostat_type_1_utc.cool_runtime_daily, pd.Series)
    assert isinstance(thermostat_type_1_utc.cool_runtime_hourly, pd.Series)
    assert thermostat_type_1_utc_bad is None


def test_too_many_minutes(thermostat_type_1_too_many_minutes):
    # None of the thermostats in this list should import
    assert len(thermostat_type_1_too_many_minutes) == 0


def test_bad_zipcode(thermostat_type_1_zip_bad):
    assert len(thermostat_type_1_zip_bad) == 0


def test_data_out_of_order(thermostat_type_1_data_out_of_order):
    assert len(thermostat_type_1_data_out_of_order) == 0


def test_data_missing_headers(thermostat_type_1_data_missing_header):
    assert len(thermostat_type_1_data_missing_header) == 0



from thermostat_nw.importers import from_csv
from thermostat_nw.importers import get_single_thermostat
from thermostat_nw.util.testing import get_data_path
from thermostat_nw.core import Thermostat, CoreDaySet
from tempfile import TemporaryDirectory

import pandas as pd
import numpy as np
from numpy import nan

import pytest

""" Abbreviations used in this file:

    XX - Heating system type: fu (furnace), hp (heat pump),  er (electric resistance), ot (other)
    XX - Cooling system type: ce (central), hp (heat pump), 

    YY - backup: eb (electric backup), ne (non electric backup, df (dual fuel), na (N/A)

    Z - speed: 1 (single stage), 2 (two stage), v (variable)
"""


@pytest.fixture(
    scope="session",
    params=[
        "../data/two_stage_ert/metadata_furnace_or_boiler_two_stage_central_two_stage.csv"
    ],
)
def thermostat_ert_fu_2_ce_2(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session",
    params=[
        "../data/two_stage_ert/metadata_furnace_or_boiler_two_stage_none_single_stage.csv"
    ],
)
def thermostat_ert_fu_2_na_1(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session",
    params=[
        "../data/two_stage_ert/metadata_heat_pump_electric_backup_two_stage_heat_pump_two_stage.csv"
    ],
)
def thermostat_ert_hpeb_2_hp_2(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session",
    params=["../data/two_stage_ert/metadata_none_two_stage_heat_pump_two_stage.csv"],
)
def thermostat_ert_na_2_hp_2(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(scope="session")
def core_heating_day_set_ert_hpeb_2_hp_2_entire(thermostat_ert_hpeb_2_hp_2):
    return thermostat_ert_hpeb_2_hp_2.get_core_heating_days(method="entire_dataset")[0]


@pytest.fixture(scope="session")
def core_cooling_day_set_ert_hpeb_2_hp_2_entire(thermostat_ert_hpeb_2_hp_2):
    return thermostat_ert_hpeb_2_hp_2.get_core_cooling_days(method="entire_dataset")[0]


@pytest.fixture(scope="session")
def metrics_ert_hpeb_2_hp_2_data():
    data = [
        {
            "sw_version": "2.0.0nw1",
            "ct_identifier": "c61badb0e0c0a7e06932de804af43111",
            "heat_type": "heat_pump_electric_backup",
            "heat_stage": "two_stage",
            "cool_type": "heat_pump",
            "cool_stage": "two_stage",
            "heating_or_cooling": "cooling_ALL",
            "zipcode": "27313",
            "station": "723170",
            "climate_zone": "Mixed-Humid",
            "start_date": "2018-01-01T00:00:00",
            "end_date": "2018-12-31T00:00:00",
            "n_days_in_inputfile_date_range": 364,
            "n_days_both_heating_and_cooling": 17,
            "n_days_insufficient_data": 51,
            "n_core_cooling_days": 159,
            "baseline_percentile_core_cooling_comfort_temperature": 70.9,
            "regional_average_baseline_cooling_comfort_temperature": 73.0,
            "percent_savings_baseline_percentile": 7.950071961514231,
            "avoided_daily_mean_core_day_runtime_baseline_percentile": 26.653874266554716,
            "avoided_total_core_day_runtime_baseline_percentile": 4237.9660083822,
            "baseline_daily_mean_core_day_runtime_baseline_percentile": 335.2658239520893,
            "baseline_total_core_day_runtime_baseline_percentile": 53307.2660083822,
            "_daily_mean_core_day_demand_baseline_baseline_percentile": 12.16814778001588,
            "percent_savings_baseline_regional": -9.462987412786953,
            "avoided_daily_mean_core_day_runtime_baseline_regional": -26.67925537512516,
            "avoided_total_core_day_runtime_baseline_regional": -4242.0016046449,
            "baseline_daily_mean_core_day_runtime_baseline_regional": 281.93269431040943,
            "baseline_total_core_day_runtime_baseline_regional": 44827.2983953551,
            "_daily_mean_core_day_demand_baseline_baseline_regional": 10.232473587517678,
            "percent_savings_baseline_hourly_regional": 1.5207727826062385,
            "avoided_daily_mean_core_day_runtime_baseline_hourly_regional": 4.765762960677582,
            "avoided_total_core_day_runtime_baseline_hourly_regional": 757.7563107477356,
            "baseline_daily_mean_core_day_runtime_baseline_hourly_regional": 313.37771264621216,
            "baseline_total_core_day_runtime_baseline_hourly_regional": 49827.056310747736,
            "_daily_mean_core_day_demand_baseline_baseline_hourly_regional": 11.373740017674411,
            "mean_demand": 11.200771275121225,
            "tau": 8.371187895647756,
            "alpha": 27.552740976955135,
            "mean_sq_err": 2714.7353166535677,
            "root_mean_sq_err": 52.103121947284194,
            "cv_root_mean_sq_err": 0.16883053945375592,
            "mean_abs_pct_err": 0.1270126045228687,
            "mean_abs_err": 39.19760751644026,
            "cov_x": [[8.890444749578084e-05]],
            "nfev": 19,
            "mesg": "Both actual and predicted relative reductions in the sum of squares  are at most 0.000000 and the relative error between two consecutive iterates is at   most 0.000000",
            "total_core_cooling_runtime": 49069.3,
            "daily_mean_core_cooling_runtime": 308.6119496855346,
            "core_cooling_days_mean_indoor_temperature": 71.87575995807128,
            "core_cooling_days_mean_outdoor_temperature": 74.44318985849057,
            "core_mean_indoor_temperature": 71.87575995807128,
            "core_mean_outdoor_temperature": 74.44318985849057,
            "heat_gain_constant": 0.05279721300559318,
            "heat_loss_constant": -0.02608637965436161,
            "hvac_constant": -0.019935339745544286,
            "overall_temperature_variance": 0.8188646995078152,
            "weekly_temperature_variance": 0.2112426460503252,
            "avg_daily_cooling_runtime": 308.6119496855346,
            "avg_daily_heating_runtime": 0.0,
            "avg_daily_auxiliary_runtime": 0.0,
            "avg_daily_emergency_runtime": 0.0,
            "lm_intercept": -190.5108328567515,
            "lm_intercept_se": 9.186306075779259,
            "lm_main_slope": 0.8169779255914287,
            "lm_main_slope_se": 0.027121972079575743,
            "lm_secondary_slope": nan,
            "lm_secondary_slope_se": nan,
            "lm_cvrmse": 0.1536874480032085,
            "lm_rsquared": 0.8524930786895475,
        },
        {
            "sw_version": "2.0.0nw1",
            "ct_identifier": "c61badb0e0c0a7e06932de804af43111",
            "heat_type": "heat_pump_electric_backup",
            "heat_stage": "two_stage",
            "cool_type": "heat_pump",
            "cool_stage": "two_stage",
            "heating_or_cooling": "heating_ALL",
            "zipcode": "27313",
            "station": "723170",
            "climate_zone": "Mixed-Humid",
            "start_date": "2018-01-01T00:00:00",
            "end_date": "2018-12-31T00:00:00",
            "n_days_in_inputfile_date_range": 364,
            "n_days_both_heating_and_cooling": 17,
            "n_days_insufficient_data": 51,
            "n_core_heating_days": 116,
            "baseline_percentile_core_heating_comfort_temperature": 70.0,
            "regional_average_baseline_heating_comfort_temperature": 69,
            "percent_savings_baseline_percentile": 9.864369103937042,
            "avoided_daily_mean_core_day_runtime_baseline_percentile": 32.91741702630587,
            "avoided_total_core_day_runtime_baseline_percentile": 3818.4203750514807,
            "baseline_daily_mean_core_day_runtime_baseline_percentile": 333.70017564699555,
            "baseline_total_core_day_runtime_baseline_percentile": 38709.22037505148,
            "_daily_mean_core_day_demand_baseline_baseline_percentile": 13.17555760521648,
            "percent_savings_baseline_regional": 3.3148699002254034,
            "avoided_daily_mean_core_day_runtime_baseline_regional": 10.312399766433282,
            "avoided_total_core_day_runtime_baseline_regional": 1196.2383729062608,
            "baseline_daily_mean_core_day_runtime_baseline_regional": 311.0951583871229,
            "baseline_total_core_day_runtime_baseline_regional": 36087.03837290626,
            "_daily_mean_core_day_demand_baseline_baseline_regional": 12.283038725066332,
            "percent_savings_baseline_hourly_regional": -5.023433690213368,
            "avoided_daily_mean_core_day_runtime_baseline_hourly_regional": -14.386905759978857,
            "avoided_total_core_day_runtime_baseline_hourly_regional": -1668.8810681575474,
            "baseline_daily_mean_core_day_runtime_baseline_hourly_regional": 286.3958528607109,
            "baseline_total_core_day_runtime_baseline_hourly_regional": 33221.91893184246,
            "_daily_mean_core_day_demand_baseline_baseline_hourly_regional": 11.307830599565262,
            "mean_demand": 11.875871971536075,
            "tau": 13.62446309141154,
            "alpha": 25.327214653509362,
            "mean_sq_err": 5950.792154485654,
            "root_mean_sq_err": 77.1413777066864,
            "cv_root_mean_sq_err": 0.2564687486092501,
            "mean_abs_pct_err": 0.1972627612697482,
            "mean_abs_err": 59.3332375078494,
            "cov_x": [[7.669596060995188e-05]],
            "nfev": 15,
            "mesg": "Both actual and predicted relative reductions in the sum of squares  are at most 0.000000 and the relative error between two consecutive iterates is at   most 0.000000",
            "total_core_heating_runtime": 34890.799999999996,
            "daily_mean_core_heating_runtime": 300.7827586206896,
            "core_heating_days_mean_indoor_temperature": 68.65607040229885,
            "core_heating_days_mean_outdoor_temperature": 43.640953728448274,
            "core_mean_indoor_temperature": 68.65607040229885,
            "core_mean_outdoor_temperature": 43.640953728448274,
            "heat_gain_constant": 0.6405944716697065,
            "heat_loss_constant": -0.003685133822820946,
            "hvac_constant": -0.001755439909246502,
            "overall_temperature_variance": 0.9840147822568611,
            "weekly_temperature_variance": 0.3322558012005256,
            "avg_daily_cooling_runtime": 0.0,
            "avg_daily_heating_runtime": 300.7827586206896,
            "avg_daily_auxiliary_runtime": 13.039655172413791,
            "avg_daily_emergency_runtime": 0.0,
            "lm_intercept": 329.1557821437414,
            "lm_intercept_se": 18.073278334062515,
            "lm_main_slope": 1.0062774143459097,
            "lm_main_slope_se": 0.06972549048205005,
            "lm_secondary_slope": 0.732217542181257,
            "lm_secondary_slope_se": 0.32923234766070597,
            "lm_cvrmse": 0.33438284149023306,
            "lm_rsquared": 0.763119673726163,
            "total_auxiliary_heating_core_day_runtime": 1512.6000000000001,
            "total_emergency_heating_core_day_runtime": 0.0,
            "dnru_daily": 0.08165462151765186,
            "dnru_reduction_daily": 0.2656491449152059,
            "mu_estimate_daily": 23.50944070120685,
            "sigma_estimate_daily": 3.1154949764320174,
            "sigmoid_model_error_daily": 0.009478437714417857,
            "sigmoid_integral_daily": 4.702030816995914,
            "aux_exceeds_heat_runtime_daily": True,
            "dnru_hourly": 0.049357181761880725,
            "dnru_reduction_hourly": 0.3344944124890115,
            "mu_estimate_hourly": 2.0833950634791125,
            "sigma_estimate_hourly": 17.7352417027378,
            "sigmoid_model_error_hourly": 0.0006156170263438765,
            "sigmoid_integral_hourly": 1.6280028084106708,
            "aux_exceeds_heat_runtime_hourly": False,
            "rhu1_00F_to_05F": nan,
            "rhu1_05F_to_10F": nan,
            "rhu1_10F_to_15F": nan,
            "rhu1_15F_to_20F": nan,
            "rhu1_20F_to_25F": 0.6045587903407809,
            "rhu1_25F_to_30F": 0.12034338820193748,
            "rhu1_30F_to_35F": 0.0771375321534741,
            "rhu1_35F_to_40F": 0.04561882032049096,
            "rhu1_40F_to_45F": 0.03743520199479428,
            "rhu1_45F_to_50F": 0.014363535770523898,
            "rhu1_50F_to_55F": 0.0,
            "rhu1_55F_to_60F": 0.015546218487394958,
            "rhu1_30F_to_45F": 0.050930833104836905,
            "rhu2_00F_to_05F": nan,
            "rhu2_05F_to_10F": nan,
            "rhu2_10F_to_15F": nan,
            "rhu2_15F_to_20F": nan,
            "rhu2_20F_to_25F": 0.6045587903407809,
            "rhu2_25F_to_30F": nan,
            "rhu2_30F_to_35F": 0.0771375321534741,
            "rhu2_35F_to_40F": 0.04561882032049096,
            "rhu2_40F_to_45F": 0.03743520199479428,
            "rhu2_45F_to_50F": 0.014363535770523898,
            "rhu2_50F_to_55F": 0.0,
            "rhu2_55F_to_60F": nan,
            "rhu2_30F_to_45F": 0.050930833104836905,
        },
    ]

    return data



import logging
import json
import zipcodes
import warnings
from pkg_resources import resource_stream
from eeweather import (
    get_isd_file_metadata,
    zcta_to_lat_long,
    rank_stations,
    select_station,
)
from eeweather.exceptions import UnrecognizedZCTAError, UnrecognizedUSAFIDError


logging.getLogger(__name__)
warnings.simplefilter("module", Warning)

# This is a JSON file with the zipcode to usaf_station mapping that were previously in eemeter.
with resource_stream(
    "thermostat_nw.resources", "zipcode_usaf_station.json"
) as zipcode_usaf_stream:
    zipcode_usaf_json = zipcode_usaf_stream.read().decode()
    zipcode_usaf = json.loads(zipcode_usaf_json)

MAX_METERS = 100000

METHOD = [
    ["high", 40000],
    ["medium", 40000],
    ["high", MAX_METERS],
    ["medium", MAX_METERS],
]


def _zip_to_lat_long(zipcode):
    """ Returns the lat / long for a zip code, or None if none is found. """
    zip_location_list = zipcodes.matching(zipcode)
    try:
        first_location = zip_location_list.pop()
        lat = float(first_location.get("lat"))
        lon = float(first_location.get("long"))
    except Exception:
        return None, None
    return lat, lon


def _get_closest_station_by_zcta_ranked(zcta):
    """Selects the nth ranked station from a list of ranked stations

    Parameters
    ----------
    zcta : string
        ZIP Code Tabulation Area (ZCTA)

    Returns
    -------
    station : string
        Station that was found
    lat : float
        latitude for the search
    lon : float
        longitude for the search
    """

    zcta = zcta.zfill(
        5
    )  # Ensure that we have 5 characters, and if not left-pad it with zeroes.
    try:
        lat, lon = zcta_to_lat_long(zcta)
    except UnrecognizedZCTAError:
        warnings.warn(
            "%s is not recognized as a ZCTA. Treating as a ZIP Code instead." % zcta
        )
        lat, lon = _zip_to_lat_long(zcta)

    station = None

    for min_quality, max_distance_meters in METHOD:
        station_ranking = rank_stations(
            lat,
            lon,
            minimum_quality=min_quality,
            max_distance_meters=max_distance_meters,
        )
        # Remove airport stations like A00008
        if len(station_ranking) > 0:
            station_ranking = station_ranking[
                station_ranking.index.str.contains("^[0-9]")
            ]

            station, _ = select_station(station_ranking)
            if station:
                break

    return station, lat, lon


def get_closest_station_by_zipcode(zipcode):
    """Look up the station by ZCTA / Zip Code from eeweather
    Searches for the closest station using ZCTA or Zip Codes
    The algorithm is as follows:

    1. Get the location of the ZCTA, and fall back to ZIP if the ZCTA isn't recognized

    2. Select the station that is the closest with the highest quality using the following method:
        * high quality stations within 40,000 meters
        * medium quality stations within 40,000 meters
        * high quality stations within 100,000 meters
        * medium quality stations within 100,000 meters

    3. If no station is returned then we log a warning message. Nothing is returned.

    4. If the station is unrecognized (:code:`UnrecognizedUSAFIDError`) then we log a warning message. Nothing is returned.

    Parameters
    ----------
    zipcode : string
        ZCTA / ZIP code to look up (referred to as zipcode in the code)

    Returns
    -------
    station : string or None
        Station that maps to the specified ZCTA / ZIP code
    """

    try:
        station, lat, lon = _get_closest_station_by_zcta_ranked(zipcode)

        if station is None:
            zipcode_mapping = zipcodes.matching(zipcode)
            warn(
                "No station found for ZCTA / ZIP %s (%s, %s)."
                % (
                    zipcode,
                    zipcode_mapping[0].get("city"),
                    zipcode_mapping[0].get("state"),
                )
            )
            return None

        isd_metadata = get_isd_file_metadata(str(station))
        if len(isd_metadata) == 0:
            logging.warning(
                "Zipcode %s mapped to station %s, but no ISD metadata was found."
                % (zipcode, station)
            )
            return None

    except UnrecognizedUSAFIDError:
        zipcode_mapping = zipcodes.matching(zipcode)
        logging.warning(
            "Closest station %s is not a recognized station for ZCTA / ZIP %s (%s, %s)."
            % (
                str(station),
                zipcode,
                zipcode_mapping[0].get("city"),
                zipcode_mapping[0].get("state"),
            )
        )
        return None

    except TypeError as e:
        logging.warning(
            "Unable to perform look-up for ZCTA / ZIP %s. Skipping." % zipcode
        )
        logging.warning(e)
        return None

    return str(station)



from thermostat_nw.importers import from_csv
from thermostat_nw.importers import get_single_thermostat
from thermostat_nw.util.testing import get_data_path
from thermostat_nw.core import Thermostat, CoreDaySet
from tempfile import TemporaryDirectory

import pandas as pd
import numpy as np
from numpy import nan

import pytest

# Note:
# The following fixtures can be quite slow without a prebuilt weather cache
# they the from_csv command fetches weather data. (This happens with builds on
# travis.)
# To speed this up, spoof the weather source.


@pytest.fixture(
    scope="session",
    params=["../data/single_stage/metadata_type_1_single_utc_offset_0.csv"],
)
def thermostat_type_1_utc(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session",
    params=["../data/single_stage/metadata_type_1_single_utc_offset_bad.csv"],
)
def thermostat_type_1_utc_bad(request):
    thermostats = from_csv(get_data_path(request.param))


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_type_1_single_bad_zip.csv"]
)
def thermostat_type_1_zip_bad(request):
    thermostats = from_csv(get_data_path(request.param))
    return list(thermostats)


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_multiple_same_key.csv"]
)
def thermostats_multiple_same_key(request):
    thermostats = from_csv(get_data_path(request.param))
    return thermostats


@pytest.fixture(
    scope="session",
    params=["../data/single_stage/metadata_type_1_single_too_many_minutes.csv"],
)
def thermostat_type_1_too_many_minutes(request):
    thermostats = from_csv(get_data_path(request.param))
    return list(thermostats)


@pytest.fixture(
    scope="session",
    params=["../data/single_stage/metadata_type_1_single_data_out_of_order.csv"],
)
def thermostat_type_1_data_out_of_order(request):
    thermostats = from_csv(get_data_path(request.param))
    return list(thermostats)


@pytest.fixture(
    scope="session",
    params=["../data/single_stage/metadata_type_1_single_data_missing_header.csv"],
)
def thermostat_type_1_data_missing_header(request):
    thermostats = from_csv(get_data_path(request.param))
    return list(thermostats)


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_type_1_single.csv"]
)
def thermostat_type_1(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_type_1_single.csv"]
)
def thermostat_type_1_cache(request):
    with TemporaryDirectory() as tempdir:
        thermostats = from_csv(
            get_data_path(request.param), save_cache=True, cache_path=tempdir
        )
        return next(thermostats)


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_type_2_single.csv"]
)
def thermostat_type_2(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_type_3_single.csv"]
)
def thermostat_type_3(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_type_4_single.csv"]
)
def thermostat_type_4(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_type_5_single.csv"]
)
def thermostat_type_5(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session", params=["../data/single_stage/metadata_single_zero_days.csv"]
)
def thermostat_zero_days(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session",
    params=["../data/single_stage/metadata_single_emg_aux_constant_on_outlier.csv"],
)
def thermostat_emg_aux_constant_on_outlier(request):
    thermostats = from_csv(get_data_path(request.param))
    return thermostats


@pytest.fixture(scope="session")
def core_heating_day_set_type_1_mid_to_mid(thermostat_type_1):
    return thermostat_type_1.get_core_heating_days(method="year_mid_to_mid")[0]


@pytest.fixture(scope="session")
def core_heating_day_set_type_1_entire(thermostat_type_1):
    return thermostat_type_1.get_core_heating_days(method="entire_dataset")[0]


@pytest.fixture(scope="session")
def core_heating_day_set_type_2(thermostat_type_2):
    return thermostat_type_2.get_core_heating_days(method="year_mid_to_mid")[0]


@pytest.fixture(scope="session")
def core_heating_day_set_type_3(thermostat_type_3):
    return thermostat_type_3.get_core_heating_days(method="year_mid_to_mid")[0]


@pytest.fixture(scope="session")
def core_heating_day_set_type_4(thermostat_type_4):
    return thermostat_type_4.get_core_heating_days(method="year_mid_to_mid")[0]


@pytest.fixture(scope="session")
def core_cooling_day_set_type_1_end_to_end(thermostat_type_1):
    return thermostat_type_1.get_core_cooling_days(method="year_end_to_end")[0]


@pytest.fixture(scope="session")
def core_cooling_day_set_type_1_entire(thermostat_type_1):
    return thermostat_type_1.get_core_cooling_days(method="entire_dataset")[0]


@pytest.fixture(scope="session")
def core_cooling_day_set_type_1_empty(thermostat_type_1):
    core_cooling_day_set = thermostat_type_1.get_core_cooling_days(
        method="entire_dataset"
    )[0]
    core_day_set = CoreDaySet(
        "empty",
        pd.Series(
            np.tile(False, core_cooling_day_set.daily.shape),
            index=core_cooling_day_set.daily.index,
        ),
        pd.Series(
            np.tile(False, core_cooling_day_set.hourly.shape),
            index=core_cooling_day_set.hourly.index,
        ),
        core_cooling_day_set.start_date,
        core_cooling_day_set.end_date,
    )
    return core_day_set


@pytest.fixture(scope="session")
def core_heating_day_set_type_1_empty(thermostat_type_1):
    core_heating_day_set = thermostat_type_1.get_core_heating_days(
        method="entire_dataset"
    )[0]
    core_day_set = CoreDaySet(
        "empty",
        pd.Series(
            np.tile(False, core_heating_day_set.daily.shape),
            index=core_heating_day_set.daily.index,
        ),
        pd.Series(
            np.tile(False, core_heating_day_set.hourly.shape),
            index=core_heating_day_set.hourly.index,
        ),
        core_heating_day_set.start_date,
        core_heating_day_set.end_date,
    )
    return core_day_set


@pytest.fixture(scope="session")
def core_cooling_day_set_type_2(thermostat_type_2):
    return thermostat_type_2.get_core_cooling_days(method="year_end_to_end")[0]


@pytest.fixture(scope="session")
def core_cooling_day_set_type_3(thermostat_type_3):
    return thermostat_type_3.get_core_cooling_days(method="year_end_to_end")[0]


@pytest.fixture(scope="session")
def core_cooling_day_set_type_5(thermostat_type_5):
    return thermostat_type_5.get_core_cooling_days(method="year_end_to_end")[0]


@pytest.fixture(scope="session")
def metrics_type_1_data():

    # this data comes from a script in scripts/test_data_generation.ipynb
    data = [
        {
            "sw_version": "2.0.0nw1",
            "ct_identifier": "8465829e-df0d-449e-97bf-96317c24dec3",
            "heat_type": "heat_pump_electric_backup",
            "heat_stage": "single_stage",
            "cool_type": "heat_pump",
            "cool_stage": "",
            "heating_or_cooling": "cooling_ALL",
            "zipcode": "62223",
            "station": "725314",
            "climate_zone": "Mixed-Humid",
            "start_date": "2011-01-01T00:00:00",
            "end_date": "2014-12-31T00:00:00",
            "n_days_in_inputfile_date_range": 1460,
            "n_days_both_heating_and_cooling": 212,
            "n_days_insufficient_data": 13,
            "n_core_cooling_days": 296,
            "baseline_percentile_core_cooling_comfort_temperature": 69.5,
            "regional_average_baseline_cooling_comfort_temperature": 73.0,
            "percent_savings_baseline_percentile": 43.83576239614899,
            "avoided_daily_mean_core_day_runtime_baseline_percentile": 192.71571072318235,
            "avoided_total_core_day_runtime_baseline_percentile": 57043.85037406198,
            "baseline_daily_mean_core_day_runtime_baseline_percentile": 439.63125126372296,
            "baseline_total_core_day_runtime_baseline_percentile": 130130.85037406199,
            "_daily_mean_core_day_demand_baseline_baseline_percentile": 9.97608214479391,
            "percent_savings_baseline_regional": 21.40352148762093,
            "avoided_daily_mean_core_day_runtime_baseline_regional": 67.24044356204358,
            "avoided_total_core_day_runtime_baseline_regional": 19903.1712943649,
            "baseline_daily_mean_core_day_runtime_baseline_regional": 314.1559841025841,
            "baseline_total_core_day_runtime_baseline_regional": 92990.1712943649,
            "_daily_mean_core_day_demand_baseline_baseline_regional": 7.128806004298179,
            "percent_savings_baseline_hourly_regional": 20.859713387979554,
            "avoided_daily_mean_core_day_runtime_baseline_hourly_regional": 65.08173810342772,
            "avoided_total_core_day_runtime_baseline_hourly_regional": 19264.194478614605,
            "baseline_daily_mean_core_day_runtime_baseline_hourly_regional": 311.9972786439683,
            "baseline_total_core_day_runtime_baseline_hourly_regional": 92351.1944786146,
            "_daily_mean_core_day_demand_baseline_baseline_hourly_regional": 7.079820808364852,
            "mean_demand": 5.602990479357406,
            "tau": -0.8165326579183203,
            "alpha": 44.06852759258279,
            "mean_sq_err": 379.2822470522948,
            "root_mean_sq_err": 19.47517001343749,
            "cv_root_mean_sq_err": 0.07887381236030344,
            "mean_abs_pct_err": 0.05040045923258922,
            "mean_abs_err": 12.444656634906245,
            "cov_x": [[1.8995481000960314e-05]],
            "nfev": 13,
            "mesg": "Both actual and predicted relative reductions in the sum of squares  are at most 0.000000 and the relative error between two consecutive iterates is at   most 0.000000",
            "total_core_cooling_runtime": 73087.0,
            "daily_mean_core_cooling_runtime": 246.91554054054055,
            "core_cooling_days_mean_indoor_temperature": 73.95971753003002,
            "core_cooling_days_mean_outdoor_temperature": 79.8426321875,
            "core_mean_indoor_temperature": 73.95971753003002,
            "core_mean_outdoor_temperature": 79.8426321875,
            "heat_gain_constant": -0.24240778006476654,
            "heat_loss_constant": -0.6987162019197792,
            "hvac_constant": 0.002734068802481422,
            "overall_temperature_variance": 3.195271934775283,
            "weekly_temperature_variance": 1.6238953447651239,
            "avg_daily_cooling_runtime": 246.91554054054055,
            "avg_daily_heating_runtime": 0.0,
            "avg_daily_auxiliary_runtime": 0.0,
            "avg_daily_emergency_runtime": 0.04310344827586207,
            "lm_intercept": -2.328160966457804,
            "lm_intercept_se": 0.6168689508193099,
            "lm_main_slope": 0.5812437420162024,
            "lm_main_slope_se": 0.002143822861219573,
            "lm_secondary_slope": nan,
            "lm_secondary_slope_se": nan,
            "lm_cvrmse": 0.02199508838237025,
            "lm_rsquared": 0.9960164062891382,
        },
        {
            "sw_version": "2.0.0nw1",
            "ct_identifier": "8465829e-df0d-449e-97bf-96317c24dec3",
            "heat_type": "heat_pump_electric_backup",
            "heat_stage": "single_stage",
            "cool_type": "heat_pump",
            "cool_stage": "",
            "heating_or_cooling": "heating_ALL",
            "zipcode": "62223",
            "station": "725314",
            "climate_zone": "Mixed-Humid",
            "start_date": "2011-01-01T00:00:00",
            "end_date": "2014-12-31T00:00:00",
            "n_days_in_inputfile_date_range": 1460,
            "n_days_both_heating_and_cooling": 212,
            "n_days_insufficient_data": 13,
            "n_core_heating_days": 895,
            "baseline_percentile_core_heating_comfort_temperature": 69.5,
            "regional_average_baseline_heating_comfort_temperature": 69,
            "percent_savings_baseline_percentile": 10.646928880334725,
            "avoided_daily_mean_core_day_runtime_baseline_percentile": 92.11668543737152,
            "avoided_total_core_day_runtime_baseline_percentile": 82444.4334664475,
            "baseline_daily_mean_core_day_runtime_baseline_percentile": 865.1948977278741,
            "baseline_total_core_day_runtime_baseline_percentile": 774349.4334664474,
            "_daily_mean_core_day_demand_baseline_baseline_percentile": 27.29246198483337,
            "percent_savings_baseline_regional": 9.033255766077659,
            "avoided_daily_mean_core_day_runtime_baseline_regional": 76.7688596268131,
            "avoided_total_core_day_runtime_baseline_regional": 68708.12936599772,
            "baseline_daily_mean_core_day_runtime_baseline_regional": 849.8470719173159,
            "baseline_total_core_day_runtime_baseline_regional": 760613.1293659977,
            "_daily_mean_core_day_demand_baseline_baseline_regional": 26.80831678982061,
            "percent_savings_baseline_hourly_regional": 6.74998510263334,
            "avoided_daily_mean_core_day_runtime_baseline_hourly_regional": 55.95995262708177,
            "avoided_total_core_day_runtime_baseline_hourly_regional": 50084.157601238185,
            "baseline_daily_mean_core_day_runtime_baseline_hourly_regional": 829.0381649175845,
            "baseline_total_core_day_runtime_baseline_hourly_regional": 741989.1576012381,
            "_daily_mean_core_day_demand_baseline_baseline_hourly_regional": 26.15190248972758,
            "mean_demand": 24.38665296761577,
            "tau": -2.348660688368403,
            "alpha": 31.70087397057362,
            "mean_sq_err": 6897.070484921533,
            "root_mean_sq_err": 83.04860314852702,
            "cv_root_mean_sq_err": 0.10742587467633805,
            "mean_abs_pct_err": 0.0719217893742982,
            "mean_abs_err": 55.60116835421654,
            "cov_x": [[4.069232611664655e-06]],
            "nfev": 15,
            "mesg": "Both actual and predicted relative reductions in the sum of squares  are at most 0.000000",
            "total_core_heating_runtime": 691905.0,
            "daily_mean_core_heating_runtime": 773.0782122905028,
            "core_heating_days_mean_indoor_temperature": 66.6958723285375,
            "core_heating_days_mean_outdoor_temperature": 44.679677289106145,
            "core_mean_indoor_temperature": 66.6958723285375,
            "core_mean_outdoor_temperature": 44.679677289106145,
            "heat_gain_constant": -0.6723011363825689,
            "heat_loss_constant": -0.21477173349444825,
            "hvac_constant": 0.03386838474491983,
            "overall_temperature_variance": 2.626759774833174,
            "weekly_temperature_variance": 0.46150361745710566,
            "avg_daily_cooling_runtime": 0.0,
            "avg_daily_heating_runtime": 773.0782122905028,
            "avg_daily_auxiliary_runtime": 167.8682634730539,
            "avg_daily_emergency_runtime": 2.2291666666666665,
            "lm_intercept": -12.76833869661695,
            "lm_intercept_se": 17.926219799986384,
            "lm_main_slope": 0.6113931244998998,
            "lm_main_slope_se": 0.033986679621240905,
            "lm_secondary_slope": 0.7075254583287997,
            "lm_secondary_slope_se": 0.05393767120226407,
            "lm_cvrmse": 0.11663640874933306,
            "lm_rsquared": 0.9412359834767653,
            "total_auxiliary_heating_core_day_runtime": 144794.0,
            "total_emergency_heating_core_day_runtime": 2104.0,
            "dnru_daily": 0.2211950234744482,
            "dnru_reduction_daily": 0.08808796490292103,
            "mu_estimate_daily": 1.7093833581698172,
            "sigma_estimate_daily": 41.02290436209435,
            "sigmoid_model_error_daily": 0.006279888996328262,
            "sigmoid_integral_daily": 3.1584001108690547,
            "aux_exceeds_heat_runtime_daily": False,
            "dnru_hourly": 0.2247928980896961,
            "dnru_reduction_hourly": 0.14563136215168795,
            "mu_estimate_hourly": -8.336263513713098,
            "sigma_estimate_hourly": 50.84672433469642,
            "sigmoid_model_error_hourly": 0.0021678686737517136,
            "sigmoid_integral_hourly": 2.8554448508840578,
            "aux_exceeds_heat_runtime_hourly": False,
            "rhu1_00F_to_05F": nan,
            "rhu1_05F_to_10F": 0.35810185185185184,
            "rhu1_10F_to_15F": 0.36336805555555557,
            "rhu1_15F_to_20F": 0.37900691389063484,
            "rhu1_20F_to_25F": 0.375673443706005,
            "rhu1_25F_to_30F": 0.3318400322150354,
            "rhu1_30F_to_35F": 0.28432496802364043,
            "rhu1_35F_to_40F": 0.19741898266605878,
            "rhu1_40F_to_45F": 0.15271363589013506,
            "rhu1_45F_to_50F": 0.09249776186213071,
            "rhu1_50F_to_55F": 0.052322643343051506,
            "rhu1_55F_to_60F": 0.028319891645631964,
            "rhu1_30F_to_45F": 0.22154695797667256,
            "rhu2_00F_to_05F": nan,
            "rhu2_05F_to_10F": 0.35810185185185184,
            "rhu2_10F_to_15F": 0.36336805555555557,
            "rhu2_15F_to_20F": 0.37900691389063484,
            "rhu2_20F_to_25F": 0.375673443706005,
            "rhu2_25F_to_30F": 0.3318400322150354,
            "rhu2_30F_to_35F": 0.28432496802364043,
            "rhu2_35F_to_40F": 0.19741898266605878,
            "rhu2_40F_to_45F": 0.15271363589013506,
            "rhu2_45F_to_50F": 0.09249776186213071,
            "rhu2_50F_to_55F": 0.052322643343051506,
            "rhu2_55F_to_60F": 0.028319891645631964,
            "rhu2_30F_to_45F": 0.22154695797667256,
        },
    ]

    return data



from thermostat_nw.parallel import schedule_batches
from thermostat_nw.equipment_type import EQUIPMENT_MAPPING

import os
import tempfile
import zipfile
from uuid import uuid4

import numpy as np
import pandas as pd
from scipy.stats import randint
import random

import pytest


@pytest.fixture
def metadata_filename():
    columns = [
        "thermostat_id",
        "heat_type",
        "heat_stage",
        "cool_type",
        "cool_stage",
        "zipcode",
        "utc_offset",
        "interval_data_filename",
    ]

    n_thermostats = 100
    thermostat_ids = [uuid4() for i in range(n_thermostats)]
    equipment_types = [random.randint(0, 5) for i in range(n_thermostats)]
    heat_type = [EQUIPMENT_MAPPING[x]["heat_type"] for x in equipment_types]
    heat_stage = [EQUIPMENT_MAPPING[x]["heat_stage"] for x in equipment_types]
    cool_type = [EQUIPMENT_MAPPING[x]["cool_type"] for x in equipment_types]
    cool_stage = [EQUIPMENT_MAPPING[x]["cool_stage"] for x in equipment_types]
    zipcodes = [
        "70754",
        "70722",
        "70726",
        "70449",
        "70442",  # "722312" 50
        "70443",
        "70441",
        "70446",
        "70447",
        "70444",  # "722312"
        "70836",
        "70778",
        "70770",
        "70774",
        "70777",  # "722312"
        "70433",
        "70437",
        "70436",
        "70435",
        "70438",  # "722312"
        "70744",
        "70748",
        "70462",
        "70465",
        "70466",  # "722312"
        "70791",
        "70714",
        "70711",
        "70451",
        "70450",  # "722312"
        "70453",
        "70455",
        "70454",
        "70456",
        "70809",  # "722312"
        "70806",
        "70807",
        "70805",
        "70769",
        "70761",  # "722312"
        "70402",
        "70403",
        "70401",
        "70737",
        "70730",  # "722312"
        "70733",
        "70739",
        "70785",
        "70789",
        "70706",  # "722312"
        "45341",
        "45344",
        "45349",
        "45319",
        "45434",  # "745700" 55
        "60018",
        "60191",
        "60193",
        "60195",
        "60194",  # "725300" 60
        "97473",
        "97449",
        "97493",
        "97467",
        "97459",  # "726917" 65
        "60421",
        "60544",
        "60404",
        "60408",
        "60481",  # "725345" 70
        "36590",
        "36564",
        "36606",
        "36605",
        "36532",  # "722235" 75
        "36541",
        "36544",
        "36568",
        "36608",
        "36609",  # "722230" 80
        "23106",
        "23060",
        "23229",
        "23222",
        "23294",  # "724029" 85
        "13674",
        "13601",
        "13606",
        "13605",
        "13682",  # "726227" 90
        "12978",
        "12972",
        "12985",
        "12903",
        "12901",  # "726225" 95
        "61051",  # "725326" 96
        "76207",  # "722589" 97
        "36362",  # "722239" 98
        "57233",  # "726546" 99
        "56289",  # "726547" 100
    ]
    utc_offsets = [-7 for _ in range(n_thermostats)]
    interval_data_filenames = ["thermostat_{}.csv".format(i) for i in thermostat_ids]

    df = pd.DataFrame(
        {
            "thermostat_id": thermostat_ids,
            "heat_type": heat_type,
            "heat_stage": heat_stage,
            "cool_type": cool_type,
            "cool_stage": cool_stage,
            "zipcode": zipcodes,
            "utc_offset": utc_offsets,
            "interval_data_filename": interval_data_filenames,
        },
        columns=columns,
    )

    temp_dir = tempfile.mkdtemp()
    metadata_filename = os.path.join(temp_dir, "metadata.csv")
    df.to_csv(metadata_filename, index=False)

    for interval_data_filename in df.interval_data_filename:
        fname = os.path.join(temp_dir, interval_data_filename)
        with open(fname, "w") as f:
            f.write("INTERVAL DATA FILE CONTENT")

    return metadata_filename


def test_schedule_batches_metadata_only(metadata_filename):

    batches = schedule_batches(metadata_filename, 5)

    assert len(batches) == 5
    assert sum([len(b) for b in batches]) == 100
    assert isinstance(batches[0], pd.DataFrame)


def test_schedule_batches_zip_files(metadata_filename):

    with pytest.raises(ValueError):
        schedule_batches(metadata_filename, 5, True)

    temp_dir = tempfile.mkdtemp()
    batch_zipfile_names = schedule_batches(metadata_filename, 5, True, temp_dir)

    assert len(batch_zipfile_names) == 5
    assert isinstance(batch_zipfile_names[0], str)

    with zipfile.ZipFile(batch_zipfile_names[0]) as zf:
        assert len(zf.infolist()) == 21



import pandas as pd
from pkg_resources import resource_stream
from collections import namedtuple


def _load_mapping(filename_or_buffer):
    df = (
        pd.read_csv(
            filename_or_buffer,
            usecols=["zipcode", "group"],
            dtype={"zipcode": str, "group": str},
        )
        .set_index("zipcode")
        .drop("zipcode", errors="ignore")
    )
    df = df.where((pd.notnull(df)), None)

    return dict(df.to_records("index"))


def retrieve_climate_zone(climate_zone_mapping, zipcode):
    """Loads the Climate Zone to Zipcode database
    and returns the climate zone and baseline regional comfort temperatures.

    Parameters
    ----------

    climate_zone_mapping : filename, default: None

        A mapping from climate zone to zipcode. If None is provided, uses
        default zipcode to climate zone mapping provided in tutorial.

        :download:`default mapping <./resources/Building America Climate Zone to Zipcode Database_Rev2_2016.09.08.csv>`

    Returns
    -------

    climate_zone_nt : named tuple
       Named Tuple consisting of the Climate Zone, baseline_regional_cooling_comfort_temperature, and baseline_regional_heating_comfort_temperature
    """

    ClimateZone = namedtuple(
        "ClimateZone",
        [
            "climate_zone",
            "baseline_regional_cooling_comfort_temperature",
            "baseline_regional_heating_comfort_temperature",
        ],
    )
    if climate_zone_mapping is None:
        with resource_stream(
            "thermostat_nw.resources",
            "Building America Climate Zone to Zipcode Database_Rev2_2016.09.08.csv",
        ) as f:
            mapping = _load_mapping(f)
    else:
        try:
            mapping = _load_mapping(climate_zone_mapping)
        except Exception as e:
            raise ValueError("Could not load climate zone mapping: %s" % e)

    with resource_stream("thermostat_nw.resources", "regional_baselines.csv") as f:
        df = pd.read_csv(
            f,
            usecols=[
                "EIA Climate Zone",
                "Baseline heating temp (F)",
                "Baseline cooling temp (F)",
            ],
        )
        df = df.where((pd.notnull(df)), None)
        df = df.set_index("EIA Climate Zone")
        cooling_regional_baseline_temps = {
            k: v for k, v in df["Baseline cooling temp (F)"].iteritems()
        }
        heating_regional_baseline_temps = {
            k: v for k, v in df["Baseline heating temp (F)"].iteritems()
        }

    climate_zone = mapping.get(zipcode)
    baseline_regional_cooling_comfort_temperature = cooling_regional_baseline_temps.get(
        climate_zone, None
    )
    baseline_regional_heating_comfort_temperature = heating_regional_baseline_temps.get(
        climate_zone, None
    )

    climate_zone_nt = ClimateZone(
        climate_zone,
        baseline_regional_cooling_comfort_temperature,
        baseline_regional_heating_comfort_temperature,
    )

    return climate_zone_nt



from thermostat_nw.importers import from_csv
from thermostat_nw.importers import get_single_thermostat
from thermostat_nw.util.testing import get_data_path
from thermostat_nw.core import Thermostat, CoreDaySet
from tempfile import TemporaryDirectory

import pandas as pd
import numpy as np
from numpy import nan

import pytest

""" Abbreviations used in this file:

    XX - Heating system type: fu (furnace), hp (heat pump),  er (electric resistance), ot (other)
    XX - Cooling system type: ce (central), hp (heat pump), 

    YY - backup: eb (electric backup), ne (non electric backup, df (dual fuel), na (N/A)

    Z - speed: 1 (single stage), 2 (two stage), v (variable)
"""


@pytest.fixture(
    scope="session",
    params=[
        "../data/two_stage/metadata_furnace_or_boiler_two_stage_central_two_stage.csv"
    ],
)
def thermostat_fu_2_ce_2(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session",
    params=[
        "../data/two_stage/metadata_furnace_or_boiler_two_stage_none_single_stage.csv"
    ],
)
def thermostat_furnace_or_boiler_two_stage_none_single_stage(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session",
    params=[
        "../data/two_stage/metadata_heat_pump_electric_backup_two_stage_heat_pump_two_stage.csv"
    ],
)
def thermostat_hpeb_2_hp_2(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(
    scope="session",
    params=["../data/two_stage/metadata_none_two_stage_heat_pump_two_stage.csv"],
)
def thermostat_na_2_hp_2(request):
    thermostats = from_csv(get_data_path(request.param))
    return next(thermostats)


@pytest.fixture(scope="session")
def core_heating_day_set_hpeb_2_hp_2_entire(thermostat_hpeb_2_hp_2):
    return thermostat_hpeb_2_hp_2.get_core_heating_days(method="entire_dataset")[0]


@pytest.fixture(scope="session")
def core_cooling_day_set_hpeb_2_hp_2_entire(thermostat_hpeb_2_hp_2):
    return thermostat_hpeb_2_hp_2.get_core_cooling_days(method="entire_dataset")[0]


@pytest.fixture(scope="session")
def metrics_hpeb_2_hp_2_data():

    data = [
        {
            "sw_version": "2.0.0nw1",
            "ct_identifier": "c61badb0e0c0a7e06932de804af43111",
            "heat_type": "heat_pump_electric_backup",
            "heat_stage": "two_stage",
            "cool_type": "heat_pump",
            "cool_stage": "two_stage",
            "heating_or_cooling": "cooling_ALL",
            "zipcode": "27313",
            "station": "723170",
            "climate_zone": "Mixed-Humid",
            "start_date": "2018-01-01T00:00:00",
            "end_date": "2018-12-31T00:00:00",
            "n_days_in_inputfile_date_range": 364,
            "n_days_both_heating_and_cooling": 17,
            "n_days_insufficient_data": 51,
            "n_core_cooling_days": 160,
            "baseline_percentile_core_cooling_comfort_temperature": 70.8,
            "regional_average_baseline_cooling_comfort_temperature": 73.0,
            "percent_savings_baseline_percentile": 8.463929746930798,
            "avoided_daily_mean_core_day_runtime_baseline_percentile": 30.116193457075433,
            "avoided_total_core_day_runtime_baseline_percentile": 4818.590953132069,
            "baseline_daily_mean_core_day_runtime_baseline_percentile": 355.8180934570754,
            "baseline_total_core_day_runtime_baseline_percentile": 56930.89495313206,
            "_daily_mean_core_day_demand_baseline_baseline_percentile": 12.482150396333545,
            "percent_savings_baseline_regional": -9.403740362536968,
            "avoided_daily_mean_core_day_runtime_baseline_regional": -27.995533727051416,
            "avoided_total_core_day_runtime_baseline_regional": -4479.285396328227,
            "baseline_daily_mean_core_day_runtime_baseline_regional": 297.7063662729486,
            "baseline_total_core_day_runtime_baseline_regional": 47633.01860367178,
            "_daily_mean_core_day_demand_baseline_baseline_regional": 10.443582566756659,
            "percent_savings_baseline_hourly_regional": 1.5955805925503548,
            "avoided_daily_mean_core_day_runtime_baseline_hourly_regional": 5.281100520953169,
            "avoided_total_core_day_runtime_baseline_hourly_regional": 844.976083352507,
            "baseline_daily_mean_core_day_runtime_baseline_hourly_regional": 330.98300052095317,
            "baseline_total_core_day_runtime_baseline_hourly_regional": 52957.2800833525,
            "_daily_mean_core_day_demand_baseline_baseline_hourly_regional": 11.61093172916648,
            "mean_demand": 11.42566995588163,
            "tau": 8.651092771281144,
            "alpha": 28.506153359728142,
            "mean_sq_err": 2864.6067628539104,
            "root_mean_sq_err": 53.52202128894153,
            "cv_root_mean_sq_err": 0.16432824398304563,
            "mean_abs_pct_err": 0.1240485579833733,
            "mean_abs_err": 40.40285102744487,
            "cov_x": [[8.205236654944263e-05]],
            "nfev": 19,
            "mesg": "Both actual and predicted relative reductions in the sum of squares  are at most 0.000000",
            "total_core_cooling_runtime": 52112.304000000004,
            "daily_mean_core_cooling_runtime": 325.7019,
            "core_cooling_days_mean_indoor_temperature": 71.86122395833334,
            "core_cooling_days_mean_outdoor_temperature": 74.372240796875,
            "core_mean_indoor_temperature": 71.86122395833334,
            "core_mean_outdoor_temperature": 74.372240796875,
            "heat_gain_constant": 0.0528959910676605,
            "heat_loss_constant": -0.024651925811247095,
            "hvac_constant": -0.02282666625140404,
            "overall_temperature_variance": 0.8432339920534964,
            "weekly_temperature_variance": 0.21954597415984842,
            "avg_daily_cooling_runtime": 325.7019,
            "avg_daily_heating_runtime": 0.0,
            "avg_daily_auxiliary_runtime": 0.0,
            "avg_daily_emergency_runtime": 0.0,
            "lm_intercept": -197.26176505508207,
            "lm_intercept_se": 8.94885392149457,
            "lm_main_slope": 0.7906805860821877,
            "lm_main_slope_se": 0.025071371277091914,
            "lm_secondary_slope": nan,
            "lm_secondary_slope_se": nan,
            "lm_cvrmse": 0.1412830071251077,
            "lm_rsquared": 0.8629179431497175,
        },
        {
            "sw_version": "2.0.0nw1",
            "ct_identifier": "c61badb0e0c0a7e06932de804af43111",
            "heat_type": "heat_pump_electric_backup",
            "heat_stage": "two_stage",
            "cool_type": "heat_pump",
            "cool_stage": "two_stage",
            "heating_or_cooling": "heating_ALL",
            "zipcode": "27313",
            "station": "723170",
            "climate_zone": "Mixed-Humid",
            "start_date": "2018-01-01T00:00:00",
            "end_date": "2018-12-31T00:00:00",
            "n_days_in_inputfile_date_range": 364,
            "n_days_both_heating_and_cooling": 17,
            "n_days_insufficient_data": 51,
            "n_core_heating_days": 116,
            "baseline_percentile_core_heating_comfort_temperature": 70.0,
            "regional_average_baseline_heating_comfort_temperature": 69,
            "percent_savings_baseline_percentile": 9.514809890645463,
            "avoided_daily_mean_core_day_runtime_baseline_percentile": 33.69972792209753,
            "avoided_total_core_day_runtime_baseline_percentile": 3909.1684389633133,
            "baseline_daily_mean_core_day_runtime_baseline_percentile": 354.18183137037346,
            "baseline_total_core_day_runtime_baseline_percentile": 41085.09243896332,
            "_daily_mean_core_day_demand_baseline_baseline_percentile": 13.71826955950154,
            "percent_savings_baseline_regional": 3.1471990461144586,
            "avoided_daily_mean_core_day_runtime_baseline_regional": 10.41395767944185,
            "avoided_total_core_day_runtime_baseline_regional": 1208.0190908152547,
            "baseline_daily_mean_core_day_runtime_baseline_regional": 330.89606112771776,
            "baseline_total_core_day_runtime_baseline_regional": 38383.94309081526,
            "_daily_mean_core_day_demand_baseline_baseline_regional": 12.816358606437076,
            "percent_savings_baseline_hourly_regional": -4.893160563800142,
            "avoided_daily_mean_core_day_runtime_baseline_hourly_regional": -14.950168166998818,
            "avoided_total_core_day_runtime_baseline_hourly_regional": -1734.2195073718628,
            "baseline_daily_mean_core_day_runtime_baseline_hourly_regional": 305.531935281277,
            "baseline_total_core_day_runtime_baseline_hourly_regional": 35441.70449262813,
            "_daily_mean_core_day_demand_baseline_baseline_hourly_regional": 11.833948203971415,
            "mean_demand": 12.41300229062868,
            "tau": 13.024414321056941,
            "alpha": 25.818258624686393,
            "mean_sq_err": 6287.039655143009,
            "root_mean_sq_err": 79.2908548014398,
            "cv_root_mean_sq_err": 0.24741117818529584,
            "mean_abs_pct_err": 0.18941422643708195,
            "mean_abs_err": 60.70386971158404,
            "cov_x": [[7.536948763697308e-05]],
            "nfev": 15,
            "mesg": "Both actual and predicted relative reductions in the sum of squares  are at most 0.000000",
            "total_core_heating_runtime": 37175.924000000006,
            "daily_mean_core_heating_runtime": 320.4821034482759,
            "core_heating_days_mean_indoor_temperature": 68.65607040229885,
            "core_heating_days_mean_outdoor_temperature": 43.640953728448274,
            "core_mean_indoor_temperature": 68.65607040229885,
            "core_mean_outdoor_temperature": 43.640953728448274,
            "heat_gain_constant": 0.6405944716697065,
            "heat_loss_constant": -0.0036649329326556234,
            "hvac_constant": -0.001968697650047877,
            "overall_temperature_variance": 0.9840147822568611,
            "weekly_temperature_variance": 0.3322558012005256,
            "avg_daily_cooling_runtime": 0.0,
            "avg_daily_heating_runtime": 320.4821034482759,
            "avg_daily_auxiliary_runtime": 13.039655172413791,
            "avg_daily_emergency_runtime": 0.0,
            "lm_intercept": 321.7662105132217,
            "lm_intercept_se": 18.46950348899804,
            "lm_main_slope": 0.9630321253417828,
            "lm_main_slope_se": 0.06652953930160889,
            "lm_secondary_slope": 0.8416804973106453,
            "lm_secondary_slope_se": 0.32430215986374933,
            "lm_cvrmse": 0.31227588928455385,
            "lm_rsquared": 0.7640382899823355,
            "total_auxiliary_heating_core_day_runtime": 1512.6000000000001,
            "total_emergency_heating_core_day_runtime": 0.0,
            "dnru_daily": 0.077526531403943,
            "dnru_reduction_daily": 0.26954443613643975,
            "mu_estimate_daily": 23.428022808877056,
            "sigma_estimate_daily": 3.0708466178842206,
            "sigmoid_model_error_daily": 0.00920383100499286,
            "sigmoid_integral_daily": 4.685775428352459,
            "aux_exceeds_heat_runtime_daily": True,
            "dnru_hourly": 0.0466793667182617,
            "dnru_reduction_hourly": 0.33630312696984904,
            "mu_estimate_hourly": 1.9960531212970736,
            "sigma_estimate_hourly": 17.6086700431372,
            "sigmoid_model_error_hourly": 0.0005936031934112763,
            "sigmoid_integral_hourly": 1.6084520686947197,
            "aux_exceeds_heat_runtime_hourly": False,
            "rhu1_00F_to_05F": nan,
            "rhu1_05F_to_10F": nan,
            "rhu1_10F_to_15F": nan,
            "rhu1_15F_to_20F": nan,
            "rhu1_20F_to_25F": 0.5963729624072203,
            "rhu1_25F_to_30F": 0.11484128153249978,
            "rhu1_30F_to_35F": 0.07377455861954366,
            "rhu1_35F_to_40F": 0.04290403220944863,
            "rhu1_40F_to_45F": 0.03502547046482063,
            "rhu1_45F_to_50F": 0.013218695583036512,
            "rhu1_50F_to_55F": 0.0,
            "rhu1_55F_to_60F": 0.014154065085746858,
            "rhu1_30F_to_45F": 0.048018860280357595,
            "rhu2_00F_to_05F": nan,
            "rhu2_05F_to_10F": nan,
            "rhu2_10F_to_15F": nan,
            "rhu2_15F_to_20F": nan,
            "rhu2_20F_to_25F": 0.5963729624072203,
            "rhu2_25F_to_30F": nan,
            "rhu2_30F_to_35F": 0.07377455861954366,
            "rhu2_35F_to_40F": 0.04290403220944863,
            "rhu2_40F_to_45F": 0.03502547046482063,
            "rhu2_45F_to_50F": 0.013218695583036512,
            "rhu2_50F_to_55F": 0.0,
            "rhu2_55F_to_60F": nan,
            "rhu2_30F_to_45F": 0.048018860280357595,
        },
    ]

    return data



import pandas as pd
import numpy as np
from numpy.testing import assert_allclose

import tempfile

import pytest

from thermostat_nw.exporters import metrics_to_csv
from thermostat_nw.multiple import (
    multiple_thermostat_calculate_epa_field_savings_metrics,
)

from .fixtures.two_stage import (
    thermostat_hpeb_2_hp_2,
    # thermostat_type_2,
    thermostat_fu_2_ce_2,
    thermostat_furnace_or_boiler_two_stage_none_single_stage,
    thermostat_na_2_hp_2,
    metrics_hpeb_2_hp_2_data,
)
from thermostat_nw.columns import EXPORT_COLUMNS
import six


@pytest.fixture(scope="session")
def metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage(
    thermostat_hpeb_2_hp_2,
):
    metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage = (
        thermostat_hpeb_2_hp_2.calculate_epa_field_savings_metrics()
    )
    return metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage


@pytest.fixture(scope="session")
def metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage_multiple(
    thermostat_hpeb_2_hp_2,
):
    metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage = (
        multiple_thermostat_calculate_epa_field_savings_metrics(
            [thermostat_hpeb_2_hp_2]
        )
    )
    return metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage


RTOL = 1e-3
ATOL = 1e-3


def test_calculate_epa_field_savings_metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage(
    metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage,
    metrics_hpeb_2_hp_2_data,
):
    assert len(metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage) == len(
        metrics_hpeb_2_hp_2_data
    )

    for key in metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage[
        0
    ].keys():
        test_value = metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage[0][
            key
        ]
        target_value = metrics_hpeb_2_hp_2_data[0][key]
        if isinstance(test_value, six.string_types):
            assert test_value == target_value
        else:
            assert_allclose(test_value, target_value, rtol=RTOL, atol=ATOL)

    for key in metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage[
        1
    ].keys():
        test_value = metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage[1][
            key
        ]
        target_value = metrics_hpeb_2_hp_2_data[1][key]
        if isinstance(test_value, six.string_types):
            assert test_value == target_value
        else:
            assert_allclose(test_value, target_value, rtol=RTOL, atol=ATOL)


def test_multiple_thermostat_calculate_epa_field_savings_metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage(
    metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage_multiple,
    metrics_hpeb_2_hp_2_data,
):
    # Test multiprocessing thermostat code
    assert len(
        metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage_multiple
    ) == len(metrics_hpeb_2_hp_2_data)

    for key in metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage_multiple[
        0
    ].keys():
        test_value = (
            metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage_multiple[0][
                key
            ]
        )
        target_value = metrics_hpeb_2_hp_2_data[0][key]
        if isinstance(test_value, six.string_types):
            assert test_value == target_value
        else:
            assert_allclose(test_value, target_value, rtol=RTOL, atol=ATOL)

    for key in metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage_multiple[
        1
    ].keys():
        test_value = (
            metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage_multiple[1][
                key
            ]
        )
        target_value = metrics_hpeb_2_hp_2_data[1][key]
        if isinstance(test_value, six.string_types):
            assert test_value == target_value
        else:
            assert_allclose(test_value, target_value, rtol=RTOL, atol=ATOL)


def test_calculate_epa_field_savings_metrics_type_3(thermostat_fu_2_ce_2):
    metrics_type_3 = thermostat_fu_2_ce_2.calculate_epa_field_savings_metrics()
    assert len(metrics_type_3) == 2


def test_calculate_epa_field_savings_metrics_type_4(
    thermostat_furnace_or_boiler_two_stage_none_single_stage,
):
    metrics_type_4 = (
        thermostat_furnace_or_boiler_two_stage_none_single_stage.calculate_epa_field_savings_metrics()
    )
    assert len(metrics_type_4) == 1


def test_calculate_epa_field_savings_metrics_type_5(thermostat_na_2_hp_2):
    metrics_type_5 = thermostat_na_2_hp_2.calculate_epa_field_savings_metrics()
    assert len(metrics_type_5) == 1


def test_metrics_to_csv(
    metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage,
):

    fd, fname = tempfile.mkstemp()
    df = metrics_to_csv(
        metrics_heat_pump_electric_backup_two_stage_heat_pump_two_stage, fname
    )

    assert isinstance(df, pd.DataFrame)
    assert df.columns[0] == "sw_version"
    assert df.columns[1] == "ct_identifier"

    with open(fname, "r") as f:
        lines = f.readlines()
        assert len(lines) == 3
        column_heads = lines[0].strip().split(",")
        assert column_heads == EXPORT_COLUMNS



import pandas as pd
import numpy as np
from numpy.testing import assert_allclose

import pytest

from .fixtures.single_stage import (
    thermostat_type_1,
    core_heating_day_set_type_1_entire as core_heating_day_set_type_1,
    core_cooling_day_set_type_1_entire as core_cooling_day_set_type_1,
    core_heating_day_set_type_1_empty,
    core_cooling_day_set_type_1_empty,
    metrics_type_1_data,
)

RTOL = 1e-3
ATOL = 1e-3


def test_get_cooling_demand_dailyavgCTD_empty(
    thermostat_type_1, core_cooling_day_set_type_1_empty, metrics_type_1_data
):
    thermostat_type_1.get_cooling_demand(core_cooling_day_set_type_1_empty)


def test_get_cooling_demand_dailyavgHTD_empty(
    thermostat_type_1, core_heating_day_set_type_1_empty, metrics_type_1_data
):
    thermostat_type_1.get_heating_demand(core_heating_day_set_type_1_empty)


def test_get_cooling_demand(
    thermostat_type_1, core_cooling_day_set_type_1, metrics_type_1_data
):
    (
        demand,
        tau_estimate,
        alpha_estimate,
        mse,
        rmse,
        cvrmse,
        mape,
        mae,
        cov_x,
        nfev,
        mesg,
    ) = thermostat_type_1.get_cooling_demand(core_cooling_day_set_type_1)
    assert_allclose(
        demand.mean(), metrics_type_1_data[0]["mean_demand"], rtol=RTOL, atol=ATOL
    )


def test_get_heating_demand(
    thermostat_type_1, core_heating_day_set_type_1, metrics_type_1_data
):
    (
        demand,
        tau_estimate,
        alpha_estimate,
        mse,
        rmse,
        cvrmse,
        mape,
        mae,
        cov_x,
        nfev,
        mesg,
    ) = thermostat_type_1.get_heating_demand(core_heating_day_set_type_1)
    assert_allclose(
        demand.mean(), metrics_type_1_data[1]["mean_demand"], rtol=RTOL, atol=ATOL
    )



from scipy.stats import norm, randint
import pandas as pd
import numpy as np
import json
from datetime import datetime

import tempfile
from itertools import islice, cycle

import pytest

from thermostat_nw.multiple import (
    multiple_thermostat_calculate_epa_field_savings_metrics,
)
from thermostat_nw.exporters import certification_to_csv
from thermostat_nw.columns import EXPORT_COLUMNS
from thermostat_nw.stats import (
    combine_output_dataframes,
    compute_summary_statistics,
    summary_statistics_to_csv,
)
from .fixtures.single_stage import thermostat_emg_aux_constant_on_outlier


def get_fake_output_df(n_columns):
    columns = EXPORT_COLUMNS

    string_placeholder = ["PLACEHOLDER"] * n_columns
    zero_column = [
        0 if randint.rvs(0, 30) > 0 else (None if randint.rvs(0, 2) > 0 else np.inf)
        for i in randint.rvs(0, 1, size=n_columns)
    ]
    one_column = [
        1 if randint.rvs(0, 30) > 0 else (None if randint.rvs(0, 2) > 0 else np.inf)
        for i in randint.rvs(0, 1, size=n_columns)
    ]
    float_column = [
        i if randint.rvs(0, 30) > 0 else (None if randint.rvs(0, 2) > 0 else np.inf)
        for i in norm.rvs(size=n_columns)
    ]
    zipcodes = ["01234", "12345", "23456", "34567", "43210", "54321", "65432", "76543"]
    zipcode_column = [i for i in islice(cycle(zipcodes), None, n_columns)]
    core_day_set_names = ["cooling_2012", "heating_2012-2013", "cooling_2013"]
    core_day_set_name_column = [
        i for i in islice(cycle(core_day_set_names), None, n_columns)
    ]

    data = {
        "sw_version": string_placeholder,
        "ct_identifier": string_placeholder,
        "heat_type": string_placeholder,
        "heat_stage": string_placeholder,
        "cool_type": string_placeholder,
        "cool_stage": string_placeholder,
        "heating_or_cooling": core_day_set_name_column,
        "station": string_placeholder,
        "zipcode": zipcode_column,
        "climate_zone": string_placeholder,
        "start_date": datetime(2011, 1, 1),
        "end_date": datetime(2012, 1, 1),
        "n_days_both_heating_and_cooling": one_column,
        "n_days_in_inputfile_date_range": one_column,
        "n_days_insufficient_data": zero_column,
        "n_core_heating_days": one_column,
        "baseline_percentile_core_cooling_comfort_temperature": float_column,
        "baseline_percentile_core_heating_comfort_temperature": float_column,
        "regional_average_baseline_cooling_comfort_temperature": float_column,
        "regional_average_baseline_heating_comfort_temperature": float_column,
        "percent_savings_baseline_percentile": float_column,
        "avoided_daily_mean_core_day_runtime_baseline_percentile": float_column,
        "avoided_total_core_day_runtime_baseline_percentile": float_column,
        "baseline_daily_mean_core_day_runtime_baseline_percentile": float_column,
        "baseline_total_core_day_runtime_baseline_percentile": float_column,
        "_daily_mean_core_day_demand_baseline_baseline_percentile": float_column,
        "percent_savings_baseline_regional": float_column,
        "avoided_daily_mean_core_day_runtime_baseline_regional": float_column,
        "avoided_total_core_day_runtime_baseline_regional": float_column,
        "baseline_daily_mean_core_day_runtime_baseline_regional": float_column,
        "baseline_total_core_day_runtime_baseline_regional": float_column,
        "_daily_mean_core_day_demand_baseline_baseline_regional": float_column,
        "percent_savings_baseline_hourly_regional": float_column,
        "avoided_daily_mean_core_day_runtime_baseline_hourly_regional": float_column,
        "avoided_total_core_day_runtime_baseline_hourly_regional": float_column,
        "baseline_daily_mean_core_day_runtime_baseline_hourly_regional": float_column,
        "baseline_total_core_day_runtime_baseline_hourly_regional": float_column,
        "_daily_mean_core_day_demand_baseline_baseline_hourly_regional": float_column,
        "mean_demand": float_column,
        "alpha": float_column,
        "tau": float_column,
        "mean_sq_err": float_column,
        "root_mean_sq_err": float_column,
        "cv_root_mean_sq_err": float_column,
        "mean_abs_err": float_column,
        "mean_abs_pct_err": float_column,
        "cov_x": string_placeholder,
        "nfev": float_column,
        "mesg": string_placeholder,
        "total_core_cooling_runtime": float_column,
        "total_core_heating_runtime": float_column,
        "total_auxiliary_heating_core_day_runtime": float_column,
        "total_emergency_heating_core_day_runtime": float_column,
        "daily_mean_core_cooling_runtime": float_column,
        "daily_mean_core_heating_runtime": float_column,
        "core_cooling_days_mean_indoor_temperature": float_column,
        "core_cooling_days_mean_outdoor_temperature": float_column,
        "core_heating_days_mean_indoor_temperature": float_column,
        "core_heating_days_mean_outdoor_temperature": float_column,
        "core_mean_indoor_temperature": float_column,
        "core_mean_outdoor_temperature": float_column,
        "heat_gain_constant": float_column,
        "heat_loss_constant": float_column,
        "hvac_constant": float_column,
        "overall_temperature_variance": float_column,
        "weekly_temperature_variance": float_column,
        "avg_daily_cooling_runtime": float_column,
        "avg_daily_heating_runtime": float_column,
        "avg_daily_auxiliary_runtime": float_column,
        "avg_daily_emergency_runtime": float_column,
        "lm_intercept": float_column,
        "lm_intercept_se": float_column,
        "lm_main_slope": float_column,
        "lm_main_slope_se": float_column,
        "lm_secondary_slope": float_column,
        "lm_secondary_slope_se": float_column,
        "lm_cvrmse": float_column,
        "lm_rsquared": float_column,
        "dnru_daily": float_column,
        "dnru_reduction_daily": float_column,
        "mu_estimate_daily": float_column,
        "sigma_estimate_daily": float_column,
        "sigmoid_model_error_daily": float_column,
        "sigmoid_integral_daily": float_column,
        "aux_exceeds_heat_runtime_daily": string_placeholder,
        "dnru_hourly": float_column,
        "dnru_reduction_hourly": float_column,
        "mu_estimate_hourly": float_column,
        "sigma_estimate_hourly": float_column,
        "sigmoid_model_error_hourly": float_column,
        "sigmoid_integral_hourly": float_column,
        "aux_exceeds_heat_runtime_hourly": string_placeholder,
        "rhu1_aux_duty_cycle": float_column,
        "rhu1_emg_duty_cycle": float_column,
        "rhu1_compressor_duty_cycle": float_column,
        "rhu1_00F_to_05F": float_column,
        "rhu1_05F_to_10F": float_column,
        "rhu1_10F_to_15F": float_column,
        "rhu1_15F_to_20F": float_column,
        "rhu1_20F_to_25F": float_column,
        "rhu1_25F_to_30F": float_column,
        "rhu1_30F_to_35F": float_column,
        "rhu1_35F_to_40F": float_column,
        "rhu1_40F_to_45F": float_column,
        "rhu1_45F_to_50F": float_column,
        "rhu1_50F_to_55F": float_column,
        "rhu1_55F_to_60F": float_column,
        "rhu1_00F_to_05F_aux_duty_cycle": float_column,
        "rhu1_05F_to_10F_aux_duty_cycle": float_column,
        "rhu1_10F_to_15F_aux_duty_cycle": float_column,
        "rhu1_15F_to_20F_aux_duty_cycle": float_column,
        "rhu1_20F_to_25F_aux_duty_cycle": float_column,
        "rhu1_25F_to_30F_aux_duty_cycle": float_column,
        "rhu1_30F_to_35F_aux_duty_cycle": float_column,
        "rhu1_35F_to_40F_aux_duty_cycle": float_column,
        "rhu1_40F_to_45F_aux_duty_cycle": float_column,
        "rhu1_45F_to_50F_aux_duty_cycle": float_column,
        "rhu1_50F_to_55F_aux_duty_cycle": float_column,
        "rhu1_55F_to_60F_aux_duty_cycle": float_column,
        "rhu1_00F_to_05F_emg_duty_cycle": float_column,
        "rhu1_05F_to_10F_emg_duty_cycle": float_column,
        "rhu1_10F_to_15F_emg_duty_cycle": float_column,
        "rhu1_15F_to_20F_emg_duty_cycle": float_column,
        "rhu1_20F_to_25F_emg_duty_cycle": float_column,
        "rhu1_25F_to_30F_emg_duty_cycle": float_column,
        "rhu1_30F_to_35F_emg_duty_cycle": float_column,
        "rhu1_35F_to_40F_emg_duty_cycle": float_column,
        "rhu1_40F_to_45F_emg_duty_cycle": float_column,
        "rhu1_45F_to_50F_emg_duty_cycle": float_column,
        "rhu1_50F_to_55F_emg_duty_cycle": float_column,
        "rhu1_55F_to_60F_emg_duty_cycle": float_column,
        "rhu1_00F_to_05F_compressor_duty_cycle": float_column,
        "rhu1_05F_to_10F_compressor_duty_cycle": float_column,
        "rhu1_10F_to_15F_compressor_duty_cycle": float_column,
        "rhu1_15F_to_20F_compressor_duty_cycle": float_column,
        "rhu1_20F_to_25F_compressor_duty_cycle": float_column,
        "rhu1_25F_to_30F_compressor_duty_cycle": float_column,
        "rhu1_30F_to_35F_compressor_duty_cycle": float_column,
        "rhu1_35F_to_40F_compressor_duty_cycle": float_column,
        "rhu1_40F_to_45F_compressor_duty_cycle": float_column,
        "rhu1_45F_to_50F_compressor_duty_cycle": float_column,
        "rhu1_50F_to_55F_compressor_duty_cycle": float_column,
        "rhu1_55F_to_60F_compressor_duty_cycle": float_column,
        "rhu2_aux_duty_cycle": float_column,
        "rhu2_emg_duty_cycle": float_column,
        "rhu2_compressor_duty_cycle": float_column,
        "rhu2_00F_to_05F": float_column,
        "rhu2_05F_to_10F": float_column,
        "rhu2_10F_to_15F": float_column,
        "rhu2_15F_to_20F": float_column,
        "rhu2_20F_to_25F": float_column,
        "rhu2_25F_to_30F": float_column,
        "rhu2_30F_to_35F": float_column,
        "rhu2_35F_to_40F": float_column,
        "rhu2_40F_to_45F": float_column,
        "rhu2_45F_to_50F": float_column,
        "rhu2_50F_to_55F": float_column,
        "rhu2_55F_to_60F": float_column,
        "rhu2_00F_to_05F_aux_duty_cycle": float_column,
        "rhu2_05F_to_10F_aux_duty_cycle": float_column,
        "rhu2_10F_to_15F_aux_duty_cycle": float_column,
        "rhu2_15F_to_20F_aux_duty_cycle": float_column,
        "rhu2_20F_to_25F_aux_duty_cycle": float_column,
        "rhu2_25F_to_30F_aux_duty_cycle": float_column,
        "rhu2_30F_to_35F_aux_duty_cycle": float_column,
        "rhu2_35F_to_40F_aux_duty_cycle": float_column,
        "rhu2_40F_to_45F_aux_duty_cycle": float_column,
        "rhu2_45F_to_50F_aux_duty_cycle": float_column,
        "rhu2_50F_to_55F_aux_duty_cycle": float_column,
        "rhu2_55F_to_60F_aux_duty_cycle": float_column,
        "rhu2_00F_to_05F_emg_duty_cycle": float_column,
        "rhu2_05F_to_10F_emg_duty_cycle": float_column,
        "rhu2_10F_to_15F_emg_duty_cycle": float_column,
        "rhu2_15F_to_20F_emg_duty_cycle": float_column,
        "rhu2_20F_to_25F_emg_duty_cycle": float_column,
        "rhu2_25F_to_30F_emg_duty_cycle": float_column,
        "rhu2_30F_to_35F_emg_duty_cycle": float_column,
        "rhu2_35F_to_40F_emg_duty_cycle": float_column,
        "rhu2_40F_to_45F_emg_duty_cycle": float_column,
        "rhu2_45F_to_50F_emg_duty_cycle": float_column,
        "rhu2_50F_to_55F_emg_duty_cycle": float_column,
        "rhu2_55F_to_60F_emg_duty_cycle": float_column,
        "rhu2_00F_to_05F_compressor_duty_cycle": float_column,
        "rhu2_05F_to_10F_compressor_duty_cycle": float_column,
        "rhu2_10F_to_15F_compressor_duty_cycle": float_column,
        "rhu2_15F_to_20F_compressor_duty_cycle": float_column,
        "rhu2_20F_to_25F_compressor_duty_cycle": float_column,
        "rhu2_25F_to_30F_compressor_duty_cycle": float_column,
        "rhu2_30F_to_35F_compressor_duty_cycle": float_column,
        "rhu2_35F_to_40F_compressor_duty_cycle": float_column,
        "rhu2_40F_to_45F_compressor_duty_cycle": float_column,
        "rhu2_45F_to_50F_compressor_duty_cycle": float_column,
        "rhu2_50F_to_55F_compressor_duty_cycle": float_column,
        "rhu2_55F_to_60F_compressor_duty_cycle": float_column,
        "rhu2_30F_to_45F": float_column,
        "rhu2_30F_to_45F_aux_duty_cycle": float_column,
        "rhu2_30F_to_45F_emg_duty_cycle": float_column,
        "rhu2_30F_to_45F_compressor_duty_cycle": float_column,
    }
    df = pd.DataFrame(data, columns=columns)
    return df


@pytest.fixture
def dataframes():
    df1 = get_fake_output_df(10)
    df2 = get_fake_output_df(10)
    dfs = [df1, df2]
    return dfs


@pytest.fixture
def combined_dataframe():
    df = get_fake_output_df(100)
    return df


def test_combine_output_dataframes(dataframes):
    combined = combine_output_dataframes(dataframes)
    assert combined.shape == (20, 119)


def test_compute_summary_statistics(combined_dataframe):
    summary_statistics = compute_summary_statistics(combined_dataframe)
    assert [len(s) for s in summary_statistics] == [
        49,
        49,
        49,
        49,
        3057,
        1573,
        3057,
        1573,
    ]

    def test_compute_summary_statistics_advanced(combined_dataframe):
        summary_statistics = compute_summary_statistics(
            combined_dataframe, advanced_filtering=True
        )
        assert [len(s) for s in summary_statistics] == [
            49,
            49,
            49,
            49,
            49,
            49,
            49,
            49,
            3057,
            1573,
            3057,
            1573,
            3057,
            1573,
            3057,
            1573,
        ]

        def test_summary_statistics_to_csv(combined_dataframe):
            summary_statistics = compute_summary_statistics(combined_dataframe)

    _, fname = tempfile.mkstemp()
    product_id = "FAKE"
    stats_df = summary_statistics_to_csv(summary_statistics, fname, product_id)
    assert isinstance(stats_df, pd.DataFrame)

    stats_df_reread = pd.read_csv(fname)
    assert stats_df_reread.shape == (3141, 5)


def test_certification(combined_dataframe):
    _, fname_stats = tempfile.mkstemp()
    _, fname_cert = tempfile.mkstemp()
    product_id = "FAKE"
    stats_df = compute_summary_statistics(combined_dataframe)
    certification_df = certification_to_csv(stats_df, fname_cert, product_id)
    assert certification_df.shape == (5, 8)


def test_iqr_filtering(thermostat_emg_aux_constant_on_outlier):

    thermostats_iqflt = list(thermostat_emg_aux_constant_on_outlier)
    # Run the metrics / statistics with the outlier thermostat in place
    iqflt_metrics = multiple_thermostat_calculate_epa_field_savings_metrics(
        thermostats_iqflt
    )
    iqflt_output_dataframe = pd.DataFrame(iqflt_metrics, columns=EXPORT_COLUMNS)
    iqflt_summary_statistics = compute_summary_statistics(iqflt_output_dataframe)

    # Remove the outlier thermostat
    thermostats_noiq = []
    for thermostat in list(thermostats_iqflt):
        if thermostat.thermostat_id != "thermostat_single_emg_aux_constant_on_outlier":
            thermostats_noiq.append(thermostat)

    if len(thermostats_noiq) == 5:
        raise ValueError("Try again")

    # Re-run the metrics / statistics with the outlier thermostat removed
    noiq_metrics = multiple_thermostat_calculate_epa_field_savings_metrics(
        thermostats_noiq
    )
    noiq_output_dataframe = pd.DataFrame(noiq_metrics, columns=EXPORT_COLUMNS)
    noiq_summary_statistics = compute_summary_statistics(noiq_output_dataframe)

    # Verify that the IQFLT removed the outliers by comparing this with the
    # metrics with the outlier thermostat already removed.
    for column in range(0, len(iqflt_summary_statistics)):
        fields_iqflt = [x for x in iqflt_summary_statistics[column] if "IQFLT" in x]
        for field_iqflt in fields_iqflt:
            field_noiq = field_iqflt.replace("rhu2IQFLT", "rhu2")
            left_side = iqflt_summary_statistics[column][field_iqflt]
            right_side = noiq_summary_statistics[column][field_noiq]

            if np.isnan(left_side) or np.isnan(right_side):
                assert np.isnan(left_side) and np.isnan(right_side)
            else:
                assert left_side == right_side



import pandas as pd
import warnings
from thermostat_nw.columns import EXPORT_COLUMNS, CERTIFICATION_HEADERS


warnings.simplefilter("module", Warning)


COLUMN_LOOKUP = {
    "percent_savings_baseline_percentile_lower_bound_95_perc_conf_national_weighted_mean": {
        "metric": "percent_savings_baseline_percentile",
        "statistic": "lower_bound_95",
    },
    "percent_savings_baseline_percentile_q20_national_weighted_mean": {
        "metric": "percent_savings_baseline_percentile",
        "statistic": "q20",
    },
    "rhu2IQFLT_30F_to_45F_upper_bound_95_perc_conf": {
        "metric": "rhu_30F_to_45F",
        "statistic": "upper_bound_95",
    },
}
FILTER_LOOKUP = {
    "national_weighted_mean_heating_tau_cvrmse_savings_p01_filter": {
        "season": "heating",
        "region": "national_weighted_mean",
        "filter": "tau_cvrmse_savings_p01",
    },
    "national_weighted_mean_cooling_tau_cvrmse_savings_p01_filter": {
        "season": "cooling",
        "region": "national_weighted_mean",
        "filter": "tau_cvrmse_savings_p01",
    },
    "all_tau_cvrmse_savings_p01_filter_heating": {
        "season": "heating",
        "region": "all",
        "filter": "tau_cvrmse_savings_p01",
    },
}

DATA_COLUMNS = [
    [
        "national_weighted_mean_heating_tau_cvrmse_savings_p01_filter",
        "percent_savings_baseline_percentile_lower_bound_95_perc_conf_national_weighted_mean",
    ],
    [
        "national_weighted_mean_cooling_tau_cvrmse_savings_p01_filter",
        "percent_savings_baseline_percentile_lower_bound_95_perc_conf_national_weighted_mean",
    ],
    [
        "national_weighted_mean_heating_tau_cvrmse_savings_p01_filter",
        "percent_savings_baseline_percentile_q20_national_weighted_mean",
    ],
    [
        "national_weighted_mean_cooling_tau_cvrmse_savings_p01_filter",
        "percent_savings_baseline_percentile_q20_national_weighted_mean",
    ],
    [
        "all_tau_cvrmse_savings_p01_filter_heating",
        "rhu2IQFLT_30F_to_45F_upper_bound_95_perc_conf",
    ],
]


def metrics_to_csv(metrics, filepath):
    """Writes metrics outputs to the file specified.

    Parameters
    ----------
    metrics : list of dict
        list of outputs from the function
        `thermostat.calculate_epa_draft_rccs_field_savings_metrics()`
    filepath : str
        filepath specification for location of output CSV file.

    Returns
    -------
    df : pd.DataFrame
        DataFrame containing data output to CSV.
    """

    output_dataframe = pd.DataFrame(metrics, columns=EXPORT_COLUMNS)
    output_dataframe.to_csv(filepath, index=False, columns=EXPORT_COLUMNS)
    return output_dataframe


def certification_to_csv(stats, filepath, product_id):
    """Writes certification outputs to the file specified.

    Parameters
    ----------
    stats : list of dict
        list of statistical outputs from the function
        `thermostat.compute_summary_statistics()`
    filepath : str
        filepath specification for location of output CSV file.

    Returns
    -------
    df : pd.DataFrame
        DataFrame containing data output to CSV.
    """
    if stats is None:
        warnings.warn("No certification data to export.")
        return None

    labels = [i.get("label") for i in stats]
    sw_version = stats[labels.index("all_tau_cvrmse_savings_p01_filter_heating")][
        "sw_version"
    ]

    certification_data = []

    for column_filter, column_data in DATA_COLUMNS:
        stats_column_number = labels.index(column_filter)
        value = stats[stats_column_number].get(column_data, None)
        row = [
            product_id,
            sw_version,
            COLUMN_LOOKUP[column_data]["metric"],
            FILTER_LOOKUP[column_filter]["filter"],
            FILTER_LOOKUP[column_filter]["region"],
            COLUMN_LOOKUP[column_data]["statistic"],
            FILTER_LOOKUP[column_filter]["season"],
            value,
        ]
        certification_data.append(row)

    output_dataframe = pd.DataFrame(certification_data, columns=CERTIFICATION_HEADERS)
    output_dataframe.to_csv(
        filepath, index=False, columns=CERTIFICATION_HEADERS, float_format="%.2f"
    )
    return output_dataframe



import pytest
import pandas as pd
from thermostat_nw.eeweather_wrapper import get_indexed_temperatures_eeweather
from .fixtures.single_stage import thermostat_type_1


def test_get_indexed_temperatures_eeweather_empty_index():
    empty_index = pd.DataFrame()
    results = get_indexed_temperatures_eeweather("720648", empty_index)
    assert results.empty is True


def test_get_index_temperatures_eeweather():
    begin_timestamp = pd.Timestamp("2011-01-01 00:00:00")
    periods = 8766
    hourly_index = pd.date_range(begin_timestamp, periods=periods, freq="H")
    results = get_indexed_temperatures_eeweather("720648", hourly_index)
    assert results.shape == (8766,)



import pytest
import warnings
import numpy as np
from numpy.testing import assert_allclose
from numpy import isnan
import pandas as pd

from datetime import datetime

from thermostat_nw.core import (
    __pandas_warnings,
    percent_savings,
    RESISTANCE_HEAT_USE_BIN_PAIRS,
)
from thermostat_nw.importers import from_csv
from thermostat_nw.util.testing import get_data_path

from .fixtures.single_stage import (
    thermostat_type_1,
    thermostat_type_2,
    thermostat_type_3,
    thermostat_type_4,
    thermostat_type_5,
    core_heating_day_set_type_1_entire,
    core_heating_day_set_type_2,
    core_heating_day_set_type_3,
    core_heating_day_set_type_4,
    core_cooling_day_set_type_1_entire,
    core_cooling_day_set_type_2,
    core_cooling_day_set_type_3,
    core_cooling_day_set_type_5,
    thermostat_zero_days,
    thermostats_multiple_same_key,
    metrics_type_1_data,
)

BASELINE_REGIONAL_COMFORT_CHECK_NONE = [
    "baseline_regional_demand",
    "baseline_regional_runtime",
    "avoided_runtime_baseline_regional",
    "percent_savings_baseline_regional",
    "avoided_daily_mean_core_day_runtime_baseline_regional",
    "avoided_total_core_day_runtime_baseline_regional",
    "baseline_daily_mean_core_day_runtime_baseline_regional",
    "baseline_total_core_day_runtime_baseline_regional",
    "_daily_mean_core_day_demand_baseline_baseline_regional",
]


def test_rhu_formatting(thermostat_type_1):
    assert "rhu1_less05F" == thermostat_type_1._format_rhu("rhu1", -np.inf, 5, None)
    assert "rhu1_greater55F" == thermostat_type_1._format_rhu("rhu1", 55, np.inf, None)
    assert "rhu1_30F_to_45F" == thermostat_type_1._format_rhu("rhu1", 30, 45, None)
    assert "rhu2_05F_to_10F_aux_duty_cycle" == thermostat_type_1._format_rhu(
        "rhu2", 5, 10, "aux_duty_cycle"
    )


def test_pandas_warnings(thermostat_type_1):
    with pytest.warns(Warning):
        __pandas_warnings("0.21.0")

    with pytest.warns(Warning):
        __pandas_warnings("1.2.0")

    with pytest.warns(None) as pytest_warnings:
        __pandas_warnings("0.25.3")
    assert not pytest_warnings

    assert __pandas_warnings(None) is None


def test_perecent_saings(thermostat_type_1):
    avoided = pd.Series([4, 5, 6])
    baseline = pd.Series([0, 0, 0])
    result = percent_savings(avoided, baseline, "thermostat_test")
    assert result == np.inf


def test_zero_days_warning(thermostat_zero_days):
    output = thermostat_zero_days.calculate_epa_field_savings_metrics()
    assert isnan(output[0]["daily_mean_core_cooling_runtime"])
    assert isnan(output[1]["daily_mean_core_heating_runtime"])


def test_multiple_same_key(thermostats_multiple_same_key):
    metrics = []
    for thermostat in thermostats_multiple_same_key:
        outputs = thermostat.calculate_epa_field_savings_metrics()
        metrics.extend(outputs)
    assert len(metrics) == 4

    for key in metrics[0]:
        if type(metrics[0][key]) == float:
            if np.isnan(metrics[0][key]):
                assert np.isnan(metrics[2][key])
            else:
                assert metrics[0][key] == metrics[2][key]
        else:
            assert metrics[0][key] == metrics[2][key]

    for key in metrics[1]:
        # nan can never be equal, so skip nan
        if not (isinstance(metrics[1][key], float) and np.isnan(metrics[1][key])):
            assert metrics[1][key] == metrics[3][key]


def test_interpolate_empty(thermostat_type_1):
    s1 = pd.Series([])
    s1_intp = thermostat_type_1._interpolate(s1)
    np.testing.assert_allclose(s1_intp, [])


def test_interpolate_full(thermostat_type_1):
    s2 = pd.Series([1])
    s2_intp = thermostat_type_1._interpolate(s2)
    np.testing.assert_allclose(s2_intp, [1])
    np.testing.assert_allclose(s2_intp, [1])


def test_interpolate_forward(thermostat_type_1):
    s3 = pd.Series([1, np.nan])
    s3_intp = thermostat_type_1._interpolate(s3)
    np.testing.assert_allclose(s3_intp, [1, 1])


def test_interpolate_backward(thermostat_type_1):
    s4 = pd.Series([np.nan, 1])
    s4_intp = thermostat_type_1._interpolate(s4)
    np.testing.assert_allclose(s4_intp, [1, 1])


def test_interpolate_one_missing_bad_method(thermostat_type_1):
    s5 = pd.Series([8, 3, np.nan, 1, 7])
    s5_intp = thermostat_type_1._interpolate(s5, method="none")
    np.testing.assert_allclose(s5_intp, [8, 3, np.nan, 1, 7])


def test_interpolate_one_missing(thermostat_type_1):
    s5 = pd.Series([8, 3, np.nan, 1, 7])
    s5_intp = thermostat_type_1._interpolate(s5)
    np.testing.assert_allclose(s5_intp, [8, 3, 2, 1, 7])


def test_interpolate_two_missing(thermostat_type_1):
    s6 = pd.Series([8, np.nan, np.nan, 1, 7])
    s6_intp = thermostat_type_1._interpolate(s6)
    np.testing.assert_allclose(s6_intp, [8, 5.666, 3.333, 1, 7], rtol=1e-3)


def test_interpolate_three_missing(thermostat_type_1):
    s = pd.Series([8, np.nan, np.nan, np.nan, 1, 7])
    s_intp = thermostat_type_1._interpolate(s)
    np.testing.assert_allclose(s_intp, [8, 6.25, np.nan, 2.75, 1, 7])


def test_thermostat_type_1_get_core_heating_days(thermostat_type_1):
    core_heating_day_sets = thermostat_type_1.get_core_heating_days(
        method="year_mid_to_mid"
    )
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_1_get_core_heating_days_bad_methods(thermostat_type_1):
    with pytest.raises(NotImplementedError) as record:
        core_heating_day_sets = thermostat_type_1.get_core_heating_days(
            method="bad_method"
        )


def test_thermostat_type_1_get_core_cooling_days(thermostat_type_1):
    core_cooling_day_sets = thermostat_type_1.get_core_cooling_days(
        method="year_end_to_end"
    )
    assert len(core_cooling_day_sets) == 4


def test_thermostat_type_1_get_core_cooling_days_bad_methods(thermostat_type_1):
    with pytest.raises(NotImplementedError) as record:
        core_cooling_day_sets = thermostat_type_1.get_core_cooling_days(
            method="bad_method"
        )


def test_thermostat_type_1_no_data(thermostat_type_1):
    heat_runtime_daily = thermostat_type_1.heat_runtime_daily
    thermostat_type_1.heat_runtime_daily = pd.Series()
    with pytest.raises(ValueError) as record:
        thermostat_type_1.validate()
    thermostat_type_1.heat_runtime_daily = heat_runtime_daily

    cool_runtime_daily = thermostat_type_1.cool_runtime_daily
    thermostat_type_1.cool_runtime_daily = pd.Series()
    with pytest.raises(ValueError) as record:
        thermostat_type_1.validate()
    thermostat_type_1.cool_runtime_daily = cool_runtime_daily

    aux_heat_runtime = thermostat_type_1.auxiliary_heat_runtime
    emg_heat_runtime = thermostat_type_1.emergency_heat_runtime
    thermostat_type_1.auxiliary_heat_runtime = None
    thermostat_type_1.emergency_heat_runtime = None
    with pytest.raises(ValueError) as record:
        thermostat_type_1.validate()
    thermostat_type_1.auxiliary_heat_runtime = aux_heat_runtime
    thermostat_type_1.emergency_heat_runtime = emg_heat_runtime


def test_thermostat_type_1_baseline_regional_comfort_temperature_none(
    thermostat_type_1,
    core_heating_day_set_type_1_entire,
    core_cooling_day_set_type_1_entire,
):
    climate_zone = "Mixed-Humid"
    core_heating_day_set_method = "entire_dataset"
    core_cooling_day_set_method = "entire_dataset"
    baseline_regional_cooling_comfort_temperature = None
    baseline_regional_heating_comfort_temperature = None
    result = thermostat_type_1._calculate_cooling_epa_field_savings_metrics(
        climate_zone,
        core_cooling_day_set_type_1_entire,
        core_cooling_day_set_method,
        baseline_regional_cooling_comfort_temperature,
    )

    for baseline_key in BASELINE_REGIONAL_COMFORT_CHECK_NONE:
        assert result.get(baseline_key, None) is None

    result = thermostat_type_1._calculate_heating_epa_field_savings_metrics(
        climate_zone,
        core_heating_day_set_type_1_entire,
        core_heating_day_set_method,
        baseline_regional_heating_comfort_temperature,
    )

    for baseline_key in BASELINE_REGIONAL_COMFORT_CHECK_NONE:
        assert result.get(baseline_key, None) is None


def test_thermostat_type_2_get_core_heating_days(thermostat_type_2):
    core_heating_day_sets = thermostat_type_2.get_core_heating_days(
        method="year_mid_to_mid"
    )
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_2_get_core_cooling_days(thermostat_type_2):
    core_cooling_day_sets = thermostat_type_2.get_core_cooling_days(
        method="year_end_to_end"
    )
    assert len(core_cooling_day_sets) == 4


def test_thermostat_type_3_get_core_heating_days(thermostat_type_3):
    core_heating_day_sets = thermostat_type_3.get_core_heating_days(
        method="year_mid_to_mid"
    )
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_3_get_core_cooling_days(thermostat_type_3):
    core_cooling_day_sets = thermostat_type_3.get_core_cooling_days(
        method="year_end_to_end"
    )
    assert len(core_cooling_day_sets) == 4


def test_thermostat_type_4_get_core_heating_days(thermostat_type_4):
    core_heating_day_sets = thermostat_type_4.get_core_heating_days(
        method="year_mid_to_mid"
    )
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_4_get_core_cooling_days(thermostat_type_4):
    with pytest.raises(ValueError):
        core_cooling_day_sets = thermostat_type_4.get_core_cooling_days(
            method="year_end_to_end"
        )


def test_thermostat_type_5_get_core_heating_days(thermostat_type_5):
    with pytest.raises(ValueError):
        core_heating_day_sets = thermostat_type_5.get_core_heating_days(
            method="year_mid_to_mid"
        )


def test_thermostat_type_5_get_core_cooling_days(thermostat_type_5):
    core_cooling_day_sets = thermostat_type_5.get_core_cooling_days(
        method="year_end_to_end"
    )
    assert len(core_cooling_day_sets) == 4


def test_thermostat_type_1_get_core_heating_days_with_params(thermostat_type_1):
    core_heating_day_sets = thermostat_type_1.get_core_heating_days(
        min_minutes_heating=0, max_minutes_cooling=0, method="year_mid_to_mid"
    )
    assert len(core_heating_day_sets) == 5


def test_thermostat_type_1_get_core_cooling_days_with_params(thermostat_type_1):
    core_heating_day_sets = thermostat_type_1.get_core_cooling_days(
        min_minutes_cooling=0, max_minutes_heating=0, method="year_end_to_end"
    )
    assert len(core_heating_day_sets) == 4


def test_get_core_cooling_day_baseline_setpoint_not_implemented(
    thermostat_type_1, core_cooling_day_set_type_1_entire
):
    with pytest.raises(NotImplementedError):
        thermostat_type_1.get_core_cooling_day_baseline_setpoint(
            core_cooling_day_set_type_1_entire, method="bad"
        )
    with pytest.raises(NotImplementedError):
        thermostat_type_1.get_core_cooling_day_baseline_setpoint(
            core_cooling_day_set_type_1_entire, source="bad"
        )
    with pytest.raises(NotImplementedError):
        thermostat_type_1.get_core_cooling_day_baseline_setpoint(
            core_cooling_day_set_type_1_entire, source="cooling_setpoint"
        )


def test_get_core_heating_day_baseline_setpoint_not_implemented(
    thermostat_type_1, core_heating_day_set_type_1_entire
):
    with pytest.raises(NotImplementedError):
        thermostat_type_1.get_core_heating_day_baseline_setpoint(
            core_heating_day_set_type_1_entire, method="bad"
        )
    with pytest.raises(NotImplementedError):
        thermostat_type_1.get_core_heating_day_baseline_setpoint(
            core_heating_day_set_type_1_entire, source="bad"
        )
    with pytest.raises(NotImplementedError):
        thermostat_type_1.get_core_heating_day_baseline_setpoint(
            core_cooling_day_set_type_1_entire, source="heating_setpoint"
        )


def test_rhu_bins_none(thermostat_type_1):
    rhu = thermostat_type_1._rhu_outputs(
        rhu_type="rhu1",
        rhu_bins=None,
        rhu_usage_bins=RESISTANCE_HEAT_USE_BIN_PAIRS,
        duty_cycle=None,
    )
    unique_rhu = set(rhu.values()).pop()
    assert unique_rhu is None


def test_thermostat_core_heating_day_set_attributes(core_heating_day_set_type_1_entire):

    assert isinstance(core_heating_day_set_type_1_entire.name, str)
    assert isinstance(core_heating_day_set_type_1_entire.daily, pd.Series)
    assert isinstance(core_heating_day_set_type_1_entire.hourly, pd.Series)
    assert core_heating_day_set_type_1_entire.daily.shape == (1461,)
    assert core_heating_day_set_type_1_entire.hourly.shape == (35064,)
    assert isinstance(
        core_heating_day_set_type_1_entire.start_date, datetime
    ) or isinstance(core_heating_day_set_type_1_entire.start_date, np.datetime64)
    assert isinstance(
        core_heating_day_set_type_1_entire.end_date, datetime
    ) or isinstance(core_heating_day_set_type_1_entire.end_date, np.datetime64)


def test_thermostat_core_cooling_day_set_attributes(core_cooling_day_set_type_1_entire):

    assert isinstance(core_cooling_day_set_type_1_entire.name, str)
    assert isinstance(core_cooling_day_set_type_1_entire.daily, pd.Series)
    assert isinstance(core_cooling_day_set_type_1_entire.hourly, pd.Series)
    assert core_cooling_day_set_type_1_entire.daily.shape == (1461,)
    assert core_cooling_day_set_type_1_entire.hourly.shape == (35064,)
    assert isinstance(
        core_cooling_day_set_type_1_entire.start_date, datetime
    ) or isinstance(core_cooling_day_set_type_1_entire.start_date, np.datetime64)
    assert isinstance(
        core_cooling_day_set_type_1_entire.end_date, datetime
    ) or isinstance(core_cooling_day_set_type_1_entire.end_date, np.datetime64)


def test_thermostat_equipment_protect_cooling(thermostat_type_4, thermostat_type_5):
    assert thermostat_type_5._protect_cooling() is None
    with pytest.raises(ValueError):
        assert thermostat_type_4._protect_cooling()


def test_thermostat_equipment_protect_heating(thermostat_type_4, thermostat_type_5):
    assert thermostat_type_4._protect_heating() is None
    with pytest.raises(ValueError):
        assert thermostat_type_5._protect_heating()


def test_thermostat_equipment_protect_resistance_heat(
    thermostat_type_1, thermostat_type_2
):
    assert thermostat_type_1._protect_resistance_heat() is None
    with pytest.raises(ValueError):
        assert thermostat_type_2._protect_resistance_heat()


def test_thermostat_equipment_protect_aux_emerg(thermostat_type_1, thermostat_type_2):
    assert thermostat_type_1._protect_aux_emerg() is None
    with pytest.raises(ValueError):
        assert thermostat_type_2._protect_aux_emerg()


def test_thermostat_type_1_total_heating_runtime(
    thermostat_type_1, core_heating_day_set_type_1_entire, metrics_type_1_data
):

    total_runtime = thermostat_type_1.total_heating_runtime(
        core_heating_day_set_type_1_entire
    )
    assert_allclose(
        total_runtime, metrics_type_1_data[1]["total_core_heating_runtime"], rtol=1e-3
    )


def test_thermostat_type_1_total_emergency_heating_runtime(
    thermostat_type_1, core_heating_day_set_type_1_entire, metrics_type_1_data
):

    total_runtime = thermostat_type_1.total_emergency_heating_runtime(
        core_heating_day_set_type_1_entire
    )
    assert_allclose(
        total_runtime,
        metrics_type_1_data[1]["total_emergency_heating_core_day_runtime"],
        rtol=1e-3,
    )


def test_thermostat_type_1_total_auxiliary_heating_runtime(
    thermostat_type_1, core_heating_day_set_type_1_entire, metrics_type_1_data
):

    total_runtime = thermostat_type_1.total_auxiliary_heating_runtime(
        core_heating_day_set_type_1_entire
    )
    assert_allclose(
        total_runtime,
        metrics_type_1_data[1]["total_auxiliary_heating_core_day_runtime"],
        rtol=1e-3,
    )


def test_thermostat_type_1_total_cooling_runtime(
    thermostat_type_1, core_cooling_day_set_type_1_entire, metrics_type_1_data
):

    total_runtime = thermostat_type_1.total_cooling_runtime(
        core_cooling_day_set_type_1_entire
    )
    assert_allclose(
        total_runtime, metrics_type_1_data[0]["total_core_cooling_runtime"], rtol=1e-3
    )


def test_thermostat_type_1_get_resistance_heat_utilization_bins_rhu1(
    thermostat_type_1, core_heating_day_set_type_1_entire, metrics_type_1_data
):

    start = 0
    stop = 60
    step = 5
    temperature_bins = list(t for t in range(start, stop + step, step))
    rhu_runtime = thermostat_type_1.get_resistance_heat_utilization_runtime(
        core_heating_day_set_type_1_entire
    )
    rhu = thermostat_type_1.get_resistance_heat_utilization_bins(
        rhu_runtime, temperature_bins, core_heating_day_set_type_1_entire
    )

    assert len(rhu) == 12

    for item in rhu.itertuples():
        bin_name = thermostat_type_1._format_rhu(
            "rhu1", item.Index.left, item.Index.right, duty_cycle=None
        )
        bin_value = item.rhu
        assert_allclose(bin_value, metrics_type_1_data[1][bin_name], rtol=1e-3)

    rhu = thermostat_type_1.get_resistance_heat_utilization_bins(
        None, temperature_bins, core_heating_day_set_type_1_entire
    )

    assert rhu is None


def test_thermostat_type_1_get_resistance_heat_utilization_bins_rhu2(
    thermostat_type_1, core_heating_day_set_type_1_entire, metrics_type_1_data
):

    start = 0
    stop = 60
    step = 5
    VAR_MIN_RHU_RUNTIME = 30 * 60  # Unit is in minutes (30 hours * 60 minutes)
    temperature_bins = list(t for t in range(start, stop + step, step))
    rhu_runtime = thermostat_type_1.get_resistance_heat_utilization_runtime(
        core_heating_day_set_type_1_entire
    )
    rhu = thermostat_type_1.get_resistance_heat_utilization_bins(
        rhu_runtime,
        temperature_bins,
        core_heating_day_set_type_1_entire,
        VAR_MIN_RHU_RUNTIME,
    )

    assert len(rhu) == 12

    for item in rhu.itertuples():
        bin_name = thermostat_type_1._format_rhu(
            "rhu2", item.Index.left, item.Index.right, duty_cycle=None
        )
        bin_value = item.rhu
        assert_allclose(bin_value, metrics_type_1_data[1][bin_name], rtol=1e-3)

    rhu = thermostat_type_1.get_resistance_heat_utilization_bins(
        None, temperature_bins, core_heating_day_set_type_1_entire, VAR_MIN_RHU_RUNTIME
    )

    assert rhu is None


def test_thermostat_types_2_get_resistance_heat_utilization_runtime_rhu(
    core_heating_day_set_type_1_entire, thermostat_type_2
):

    with pytest.raises(ValueError):
        thermostat_type_2.get_resistance_heat_utilization_runtime(
            core_heating_day_set_type_1_entire
        )


@pytest.fixture(params=range(2))
def core_days(
    request,
    thermostat_type_1,
    core_heating_day_set_type_1_entire,
    core_cooling_day_set_type_1_entire,
):

    tests = [
        (thermostat_type_1, core_cooling_day_set_type_1_entire, 0, "cooling"),
        (thermostat_type_1, core_heating_day_set_type_1_entire, 1, "heating"),
    ]

    return tests[request.param]


def test_day_counts(core_days, metrics_type_1_data):
    thermostat, core_day_set, i, heating_or_cooling = core_days
    n_both, n_days_insufficient = thermostat.get_ignored_days(core_day_set)
    n_core_days = thermostat.get_core_day_set_n_days(core_day_set)
    n_days_in_inputfile_date_range = thermostat.get_inputfile_date_range(core_day_set)
    assert n_both == metrics_type_1_data[i]["n_days_both_heating_and_cooling"]
    assert n_days_insufficient == metrics_type_1_data[i]["n_days_insufficient_data"]
    assert (
        n_core_days
        == metrics_type_1_data[i]["n_core_{}_days".format(heating_or_cooling)]
    )
    assert (
        n_days_in_inputfile_date_range
        == metrics_type_1_data[i]["n_days_in_inputfile_date_range"]
    )



from thermostat_nw.importers import from_csv
from thermostat_nw.importers import get_single_thermostat
from thermostat_nw.util.testing import get_data_path
from thermostat_nw.core import Thermostat, CoreDaySet
from tempfile import TemporaryDirectory

import pandas as pd
import numpy as np
from numpy import nan

import pytest

# will be modified, recreate every time by scoping to function
@pytest.fixture(scope="function")
def thermostat_template():
    thermostat_id = "FAKE"
    heat_type = None
    heat_stage = None
    cool_type = None
    cool_stage = None
    zipcode = "FAKE"
    station = "FAKE"
    temperature_in = pd.Series([], dtype="Float64")
    temperature_out = pd.Series([], dtype="Float64")
    cool_runtime = pd.Series([], dtype="Float64")
    heat_runtime = pd.Series([], dtype="Float64")
    auxiliary_heat_runtime = pd.Series([], dtype="Float64")
    emergency_heat_runtime = pd.Series([], dtype="Float64")

    thermostat = Thermostat(
        thermostat_id,
        heat_type,
        heat_stage,
        cool_type,
        cool_stage,
        zipcode,
        station,
        temperature_in,
        temperature_out,
        cool_runtime,
        heat_runtime,
        auxiliary_heat_runtime,
        emergency_heat_runtime,
    )
    return thermostat



import datetime
import pandas as pd
import pytest

from thermostat_nw.importers import (
    from_csv,
    normalize_utc_offset,
)

from thermostat_nw.util.testing import get_data_path

from .fixtures.two_stage import (
    thermostat_hpeb_2_hp_2,
)


def test_import_csv(thermostat_hpeb_2_hp_2):
    def assert_is_series_with_shape(series, shape):
        assert isinstance(series, pd.Series)
        assert series.shape == shape

    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.cool_runtime_daily, (365,))
    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.heat_runtime_daily, (365,))

    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.cool_runtime_hourly, (8760,))
    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.heat_runtime_hourly, (8760,))

    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.auxiliary_heat_runtime, (8760,))
    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.emergency_heat_runtime, (8760,))

    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.temperature_in, (8760,))
    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.temperature_out, (8760,))



import pandas as pd
import numpy as np

from collections import OrderedDict
from itertools import chain
import warnings
from functools import reduce
from pkg_resources import resource_stream
import logging

from thermostat_nw import get_version
from thermostat_nw.columns import (
    REAL_OR_INTEGER_VALUED_COLUMNS_HEATING,
    REAL_OR_INTEGER_VALUED_COLUMNS_COOLING,
    REAL_OR_INTEGER_VALUED_COLUMNS_ALL,
)

QUANTILE = [
    1,
    2.5,
    5,
    10,
    15,
    20,
    25,
    30,
    35,
    40,
    45,
    50,
    55,
    60,
    65,
    70,
    75,
    80,
    85,
    90,
    95,
    98,
    99,
]
IQR_FILTER_PARAMETER = 1.5
TOP_ONLY_PERCENTILE_FILTER = 0.05  # Filters top 5 percent for RHU2 calculation
UNFILTERED_PERCENTILE = 1 - TOP_ONLY_PERCENTILE_FILTER

logger = logging.getLogger("epathermostat")
warnings.simplefilter("module", Warning)


def combine_output_dataframes(dfs):
    """Combines output dataframes. Useful when combining output from batches.

    Parameters
    ----------
    dfs : list of pd.DataFrame
        Output dataFrames to combine into one.

    Returns
    -------
    out : pd.DataFrame
        Dataframe with combined output metadata.

    """
    return pd.concat(dfs, ignore_index=True)


def get_filtered_stats(
    df, row_filter, label, heating_or_cooling, target_columns, target_baseline_method
):

    n_rows_total = df.shape[0]

    filtered_df = df[[row_filter(row, df) for i, row in df.iterrows()]]

    n_rows_kept = filtered_df.shape[0]
    n_rows_discarded = n_rows_total - n_rows_kept

    stats = OrderedDict()
    stats["label"] = "{}_{}".format(label, heating_or_cooling)
    stats["sw_version"] = get_version()
    stats["n_thermostat_core_day_sets_total"] = n_rows_total
    stats["n_thermostat_core_day_sets_kept"] = n_rows_kept
    stats["n_thermostat_core_day_sets_discarded"] = n_rows_discarded

    if n_rows_total == 0:
        warnings.warn(
            "Not enough data to compute summary_statistics ({}_{})".format(
                label, heating_or_cooling
            )
        )
        return []

    for column_name in target_columns:
        column = filtered_df[column_name].replace([np.inf, -np.inf], np.nan).dropna()

        # calculate quantiles and statistics
        mean = np.nanmean(pd.to_numeric(column))

        if column.count() != 0:
            sem = np.nanstd(column) / (column.count() ** 0.5)
        else:
            sem = np.nan
        lower_bound = mean - (1.96 * sem)
        upper_bound = mean + (1.96 * sem)
        stats["{}_n".format(column_name)] = column.count()
        stats["{}_upper_bound_95_perc_conf".format(column_name)] = upper_bound
        stats["{}_mean".format(column_name)] = mean
        stats["{}_lower_bound_95_perc_conf".format(column_name)] = lower_bound
        stats["{}_sem".format(column_name)] = sem

        for quantile in QUANTILE:
            stats["{}_q{}".format(column_name, quantile)] = column.quantile(
                quantile / 100.0
            )

        # Calculate IQR for RHU2 and filter outliers
        # This adds an rhu2IQFLT version of the rhu2 column to the output data.
        if "rhu2_" in column_name:
            iqr_column_name = column_name.replace("rhu2", "rhu2IQFLT")
            iqr_filter = column < column.quantile(UNFILTERED_PERCENTILE)
            if bool(iqr_filter.any()) is False:
                iqr_filter = column == column
                warnings.warn(
                    "RHU filtering 5% and min Runtime filtering removed entire dataset from statistics summary for bin. Disabling filter."
                )
            iqr_filtered_column = column.loc[iqr_filter]

            # calculate quantiles and statistics for RHU2 IQR (IQFLT) and
            # non-IQR filtering (NOIQ)
            iqr_mean = np.nanmean(pd.to_numeric(iqr_filtered_column))
            iqr_sem = np.nanstd(iqr_filtered_column) / (
                iqr_filtered_column.count() ** 0.5
            )
            iqr_lower_bound = iqr_mean - (1.96 * iqr_sem)
            iqr_upper_bound = iqr_mean + (1.96 * iqr_sem)

            stats["{}_n".format(iqr_column_name)] = iqr_filtered_column.count()
            stats[
                "{}_upper_bound_95_perc_conf".format(iqr_column_name)
            ] = iqr_upper_bound
            stats["{}_mean".format(iqr_column_name)] = iqr_mean
            stats[
                "{}_lower_bound_95_perc_conf".format(iqr_column_name)
            ] = iqr_lower_bound
            stats["{}_sem".format(iqr_column_name)] = iqr_sem

            for quantile in QUANTILE:
                stats[
                    "{}_q{}".format(iqr_column_name, quantile)
                ] = iqr_filtered_column.quantile(quantile / 100.0)

    return [stats]


def compute_summary_statistics(
    metrics_df, target_baseline_method="baseline_percentile", advanced_filtering=False
):
    """Computes summary statistics for the output dataframe. Computes the
    following statistics for each real-valued or integer valued column in
    the output dataframe: mean, standard error of the mean, and deciles.

    Parameters
    ----------
    df : pd.DataFrame
        Output for which to compute summary statistics.
    label : str
        Name for this set of thermostat outputs.
    target_baseline_method : {"baseline_percentile", "baseline_regional"}, default "baseline_percentile"
        Baselining method by which samples will be filtered according to bad fits.

    Returns
    -------
    stats : collections.OrderedDict
        An ordered dict containing the summary statistics. Column names are as
        follows, in which ### is a placeholder for the name of the column:

          - mean: ###_mean
          - standard error of the mean: ###_sem
          - 1st quantile:  ###_1q
          - 2.5th quantile:###_2.5q
          - 5th quantile:  ###_5q
          - 10th quantile: ###_10q
          - 15th quantile: ###_15q
          - 20th quantile: ###_20q
          - 25th quantile: ###_25q
          - 30th quantile: ###_30q
          - 35th quantile: ###_35q
          - 40th quantile: ###_40q
          - 45th quantile: ###_45q
          - 50th quantile: ###_50q
          - 55th quantile: ###_55q
          - 60th quantile: ###_60q
          - 65th quantile: ###_65q
          - 70th quantile: ###_70q
          - 75th quantile: ###_75q
          - 80th quantile: ###_80q
          - 85th quantile: ###_85q
          - 90th quantile: ###_90q
          - 95th quantile: ###_95q
          - 98th quantile: ###_98q
          - 99th quantile: ###_99q
          - number of non-null core day sets: ###_n

        The following general values are also output:

          - label: label
          - number of total core day sets: n_total_core_day_sets

    """

    if target_baseline_method not in ["baseline_percentile", "baseline_regional"]:
        message = (
            "Baseline method not supported - please use one of"
            ' "baseline_percentile" or "baseline_regional"'
        )
        raise ValueError(message)

    def _identity_filter(row, df):
        return True

    def _range_filter(
        row,
        column_name,
        heating_or_cooling,
        lower_bound=-np.inf,
        upper_bound=np.inf,
        target_baseline=False,
    ):
        if target_baseline:
            full_column_selector = "{}_{}".format(column_name, target_baseline_method)
        else:
            full_column_selector = column_name
        column_value = row[full_column_selector]
        return lower_bound < column_value < upper_bound

    def _percentile_range_filter(
        row, column_name, heating_or_cooling, df, quantile=0.0, target_baseline=False
    ):
        if target_baseline:
            full_column_selector = "{}_{}".format(column_name, target_baseline_method)
        else:
            full_column_selector = column_name
        lower_bound = df[full_column_selector].dropna().quantile(0.0 + quantile)
        upper_bound = df[full_column_selector].dropna().quantile(1.0 - quantile)
        return _range_filter(
            row,
            column_name,
            heating_or_cooling,
            lower_bound,
            upper_bound,
            target_baseline,
        )

    def _tau_filter_heating(row, df):
        return _range_filter(row, "tau", "heating", 0, 25)

    def _tau_filter_cooling(row, df):
        return _range_filter(row, "tau", "cooling", 0, 25)

    def _cvrmse_filter_heating(row, df):
        return _range_filter(row, "cv_root_mean_sq_err", "heating", upper_bound=0.6)

    def _cvrmse_filter_cooling(row, df):
        return _range_filter(row, "cv_root_mean_sq_err", "cooling", upper_bound=0.6)

    def _savings_filter_p01_heating(row, df):
        return _percentile_range_filter(
            row, "percent_savings", "heating", df, 0.01, True
        )

    def _savings_filter_p01_cooling(row, df):
        return _percentile_range_filter(
            row, "percent_savings", "cooling", df, 0.01, True
        )

    def _combine_filters(filters):
        def _new_filter(row, df):
            return reduce(lambda x, y: x and y(row, df), filters, True)

        return _new_filter

    def heating_stats(df, filter_, label):
        heating_df = df[["heating" in name for name in df["heating_or_cooling"]]]
        return get_filtered_stats(
            heating_df,
            filter_,
            label,
            "heating",
            REAL_OR_INTEGER_VALUED_COLUMNS_HEATING,
            target_baseline_method,
        )

    def cooling_stats(df, filter_, label):
        cooling_df = df[["cooling" in name for name in df["heating_or_cooling"]]]
        return get_filtered_stats(
            cooling_df,
            filter_,
            label,
            "cooling",
            REAL_OR_INTEGER_VALUED_COLUMNS_COOLING,
            target_baseline_method,
        )

    if len(metrics_df) == 0:
        warnings.warn("No data to compute for summary statistics.")
        return None

    very_cold_cold_df = metrics_df[
        [
            (cz is not None) and "Very-Cold/Cold" in cz
            for cz in metrics_df["climate_zone"]
        ]
    ]
    mixed_humid_df = metrics_df[
        [(cz is not None) and "Mixed-Humid" in cz for cz in metrics_df["climate_zone"]]
    ]
    mixed_dry_hot_dry_df = metrics_df[
        [
            (cz is not None) and "Mixed-Dry/Hot-Dry" in cz
            for cz in metrics_df["climate_zone"]
        ]
    ]
    hot_humid_df = metrics_df[
        [(cz is not None) and "Hot-Humid" in cz for cz in metrics_df["climate_zone"]]
    ]
    marine_df = metrics_df[
        [(cz is not None) and "Marine" in cz for cz in metrics_df["climate_zone"]]
    ]

    filter_0 = _identity_filter
    filter_1_heating = _combine_filters([_tau_filter_heating])
    filter_1_cooling = _combine_filters([_tau_filter_cooling])
    filter_2_heating = _combine_filters([_tau_filter_heating, _cvrmse_filter_heating])
    filter_2_cooling = _combine_filters([_tau_filter_cooling, _cvrmse_filter_cooling])
    filter_3_heating = _combine_filters(
        [_tau_filter_heating, _cvrmse_filter_heating, _savings_filter_p01_heating]
    )
    filter_3_cooling = _combine_filters(
        [_tau_filter_cooling, _cvrmse_filter_cooling, _savings_filter_p01_cooling]
    )

    if advanced_filtering:
        stats = list(
            chain.from_iterable(
                [
                    heating_stats(metrics_df, filter_0, "all_no_filter"),
                    cooling_stats(metrics_df, filter_0, "all_no_filter"),
                    heating_stats(
                        very_cold_cold_df, filter_0, "very-cold_cold_no_filter"
                    ),
                    cooling_stats(
                        very_cold_cold_df, filter_0, "very-cold_cold_no_filter"
                    ),
                    heating_stats(mixed_humid_df, filter_0, "mixed-humid_no_filter"),
                    cooling_stats(mixed_humid_df, filter_0, "mixed-humid_no_filter"),
                    heating_stats(
                        mixed_dry_hot_dry_df, filter_0, "mixed-dry_hot-dry_no_filter"
                    ),
                    cooling_stats(
                        mixed_dry_hot_dry_df, filter_0, "mixed-dry_hot-dry_no_filter"
                    ),
                    heating_stats(hot_humid_df, filter_0, "hot-humid_no_filter"),
                    cooling_stats(hot_humid_df, filter_0, "hot-humid_no_filter"),
                    heating_stats(marine_df, filter_0, "marine_no_filter"),
                    cooling_stats(marine_df, filter_0, "marine_no_filter"),
                    heating_stats(metrics_df, filter_1_heating, "all_tau_filter"),
                    cooling_stats(metrics_df, filter_1_cooling, "all_tau_filter"),
                    heating_stats(
                        very_cold_cold_df, filter_1_heating, "very-cold_cold_tau_filter"
                    ),
                    cooling_stats(
                        very_cold_cold_df, filter_1_cooling, "very-cold_cold_tau_filter"
                    ),
                    heating_stats(
                        mixed_humid_df, filter_1_heating, "mixed-humid_tau_filter"
                    ),
                    cooling_stats(
                        mixed_humid_df, filter_1_cooling, "mixed-humid_tau_filter"
                    ),
                    heating_stats(
                        mixed_dry_hot_dry_df,
                        filter_1_heating,
                        "mixed-dry_hot-dry_tau_filter",
                    ),
                    cooling_stats(
                        mixed_dry_hot_dry_df,
                        filter_1_cooling,
                        "mixed-dry_hot-dry_tau_filter",
                    ),
                    heating_stats(
                        hot_humid_df, filter_1_heating, "hot-humid_tau_filter"
                    ),
                    cooling_stats(
                        hot_humid_df, filter_1_cooling, "hot-humid_tau_filter"
                    ),
                    heating_stats(marine_df, filter_1_heating, "marine_tau_filter"),
                    cooling_stats(marine_df, filter_1_cooling, "marine_tau_filter"),
                    heating_stats(
                        metrics_df, filter_2_heating, "all_tau_cvrmse_filter"
                    ),
                    cooling_stats(
                        metrics_df, filter_2_cooling, "all_tau_cvrmse_filter"
                    ),
                    heating_stats(
                        very_cold_cold_df,
                        filter_2_heating,
                        "very-cold_cold_tau_cvrmse_filter",
                    ),
                    cooling_stats(
                        very_cold_cold_df,
                        filter_2_cooling,
                        "very-cold_cold_tau_cvrmse_filter",
                    ),
                    heating_stats(
                        mixed_humid_df,
                        filter_2_heating,
                        "mixed-humid_tau_cvrmse_filter",
                    ),
                    cooling_stats(
                        mixed_humid_df,
                        filter_2_cooling,
                        "mixed-humid_tau_cvrmse_filter",
                    ),
                    heating_stats(
                        mixed_dry_hot_dry_df,
                        filter_2_heating,
                        "mixed-dry_hot-dry_tau_cvrmse_filter",
                    ),
                    cooling_stats(
                        mixed_dry_hot_dry_df,
                        filter_2_cooling,
                        "mixed-dry_hot-dry_tau_cvrmse_filter",
                    ),
                    heating_stats(
                        hot_humid_df, filter_2_heating, "hot-humid_tau_cvrmse_filter"
                    ),
                    cooling_stats(
                        hot_humid_df, filter_2_cooling, "hot-humid_tau_cvrmse_filter"
                    ),
                    heating_stats(
                        marine_df, filter_2_heating, "marine_tau_cvrmse_filter"
                    ),
                    cooling_stats(
                        marine_df, filter_2_cooling, "marine_tau_cvrmse_filter"
                    ),
                    heating_stats(
                        metrics_df,
                        filter_3_heating,
                        "all_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        metrics_df,
                        filter_3_cooling,
                        "all_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        very_cold_cold_df,
                        filter_3_heating,
                        "very-cold_cold_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        very_cold_cold_df,
                        filter_3_cooling,
                        "very-cold_cold_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        mixed_humid_df,
                        filter_3_heating,
                        "mixed-humid_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        mixed_humid_df,
                        filter_3_cooling,
                        "mixed-humid_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        mixed_dry_hot_dry_df,
                        filter_3_heating,
                        "mixed-dry_hot-dry_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        mixed_dry_hot_dry_df,
                        filter_3_cooling,
                        "mixed-dry_hot-dry_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        hot_humid_df,
                        filter_3_heating,
                        "hot-humid_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        hot_humid_df,
                        filter_3_cooling,
                        "hot-humid_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        marine_df,
                        filter_3_heating,
                        "marine_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        marine_df,
                        filter_3_cooling,
                        "marine_tau_cvrmse_savings_p01_filter",
                    ),
                ]
            )
        )
    else:
        stats = list(
            chain.from_iterable(
                [
                    heating_stats(metrics_df, filter_0, "all_no_filter"),
                    cooling_stats(metrics_df, filter_0, "all_no_filter"),
                    heating_stats(
                        very_cold_cold_df, filter_0, "very-cold_cold_no_filter"
                    ),
                    cooling_stats(
                        very_cold_cold_df, filter_0, "very-cold_cold_no_filter"
                    ),
                    heating_stats(mixed_humid_df, filter_0, "mixed-humid_no_filter"),
                    cooling_stats(mixed_humid_df, filter_0, "mixed-humid_no_filter"),
                    heating_stats(
                        mixed_dry_hot_dry_df, filter_0, "mixed-dry_hot-dry_no_filter"
                    ),
                    cooling_stats(
                        mixed_dry_hot_dry_df, filter_0, "mixed-dry_hot-dry_no_filter"
                    ),
                    heating_stats(hot_humid_df, filter_0, "hot-humid_no_filter"),
                    cooling_stats(hot_humid_df, filter_0, "hot-humid_no_filter"),
                    heating_stats(marine_df, filter_0, "marine_no_filter"),
                    cooling_stats(marine_df, filter_0, "marine_no_filter"),
                    heating_stats(
                        metrics_df,
                        filter_3_heating,
                        "all_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        metrics_df,
                        filter_3_cooling,
                        "all_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        very_cold_cold_df,
                        filter_3_heating,
                        "very-cold_cold_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        very_cold_cold_df,
                        filter_3_cooling,
                        "very-cold_cold_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        mixed_humid_df,
                        filter_3_heating,
                        "mixed-humid_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        mixed_humid_df,
                        filter_3_cooling,
                        "mixed-humid_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        mixed_dry_hot_dry_df,
                        filter_3_heating,
                        "mixed-dry_hot-dry_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        mixed_dry_hot_dry_df,
                        filter_3_cooling,
                        "mixed-dry_hot-dry_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        hot_humid_df,
                        filter_3_heating,
                        "hot-humid_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        hot_humid_df,
                        filter_3_cooling,
                        "hot-humid_tau_cvrmse_savings_p01_filter",
                    ),
                    heating_stats(
                        marine_df,
                        filter_3_heating,
                        "marine_tau_cvrmse_savings_p01_filter",
                    ),
                    cooling_stats(
                        marine_df,
                        filter_3_cooling,
                        "marine_tau_cvrmse_savings_p01_filter",
                    ),
                ]
            )
        )

    stats_dict = {stat["label"]: stat for stat in stats}

    def _load_climate_zone_weights(filename_or_buffer):
        climate_zone_keys = {
            "Very-Cold/Cold": "very-cold_cold",
            "Mixed-Humid": "mixed-humid",
            "Mixed-Dry/Hot-Dry": "mixed-dry_hot-dry",
            "Hot-Humid": "hot-humid",
            "Marine": "marine",
        }
        df = pd.read_csv(
            filename_or_buffer,
            usecols=["climate_zone", "heating_weight", "cooling_weight"],
        ).set_index("climate_zone")

        heating_weights = {
            climate_zone_keys[cz]: weight
            for cz, weight in df["heating_weight"].iteritems()
        }
        cooling_weights = {
            climate_zone_keys[cz]: weight
            for cz, weight in df["cooling_weight"].iteritems()
        }

        return heating_weights, cooling_weights

    with resource_stream(
        "thermostat_nw.resources", "NationalAverageClimateZoneWeightings.csv"
    ) as f:
        heating_weights, cooling_weights = _load_climate_zone_weights(f)

    def _compute_national_weightings(stats_by_climate_zone, keys, weights):
        def _national_weight(key):
            results = []
            for cz, weight in weights.items():
                stat_cz = stats_by_climate_zone.get(cz)
                if stat_cz is None:
                    value = None
                else:
                    value = stat_cz.get(key)
                if pd.notnull(weight) and pd.notnull(value):
                    results.append((weight, value))
            if len(results) == 0:
                return None
            else:
                weighted_sum = sum(
                    [weight * value for weight, value in results]
                )  # noqa: F812
                sum_of_weights = sum([weight for weight, _ in results])
                return weighted_sum / sum_of_weights

        stats = [
            "mean",
            "q1",
            "q2.5",
            "q5",
            "q10",
            "q15",
            "q20",
            "q25",
            "q30",
            "q35",
            "q40",
            "q45",
            "q50",
            "q55",
            "q60",
            "q65",
            "q70",
            "q75",
            "q80",
            "q85",
            "q90",
            "q95",
            "q98",
            "q99",
        ]

        key_stats = ["{}_{}".format(key, stat) for key in keys for stat in stats]

        return {
            "{}_{}".format(key_stat, "national_weighted_mean"): _national_weight(
                key_stat
            )
            for key_stat in key_stats
        }

    def _compute_national_weighting_lower_and_upper_bounds(
        stats_by_climate_zone, keys, weights
    ):
        def _compute_bounds(key):

            # compute sem savings
            means, sems, weights_ = [], [], []
            for cz, weight in weights.items():
                stat_cz = stats_by_climate_zone.get(cz)
                if stat_cz is None:
                    mean, sem = None, None
                else:
                    mean = stat_cz.get("{}_mean".format(key), None)
                    sem = stat_cz.get("{}_sem".format(key), None)

                if pd.notnull(weight) and pd.notnull(mean) and pd.notnull(sem):
                    weights_.append(weight)
                    means.append(mean)
                    sems.append(sem)

            if len(weights_) == 0:
                return {}
            else:
                weighted_sum = sum(
                    [
                        weight * mean
                        for weight, mean in zip(weights_, means)  # noqa: F812
                    ]
                )
                weighted_mean = weighted_sum / sum(weights_)  # renormalize

                weighted_sem = (
                    sum(
                        [
                            (weight * sem) ** 2
                            for weight, sem in zip(weights_, sems)  # noqa: F812
                        ]
                    )
                    ** 0.5
                )

                lower_bound = weighted_mean - (1.96 * weighted_sem)
                upper_bound = weighted_mean + (1.96 * weighted_sem)

                return {
                    "{}_lower_bound_95_perc_conf_national_weighted_mean".format(
                        key
                    ): lower_bound,
                    "{}_upper_bound_95_perc_conf_national_weighted_mean".format(
                        key
                    ): upper_bound,
                }

        items = {}
        for key in keys:
            items.update(_compute_bounds(key))
        return items

    national_weighting_stats = []

    if advanced_filtering:
        filters = [
            "no_filter",
            "tau_filter",
            "tau_cvrmse_filter",
            "tau_cvrmse_savings_p01_filter",
        ]
    else:
        filters = [
            "no_filter",
            "tau_cvrmse_savings_p01_filter",
        ]

    climate_zones = [
        "mixed-humid",
        "mixed-dry_hot-dry",
        "marine",
        "hot-humid",
        "very-cold_cold",
    ]
    methods = [
        "baseline_percentile",
        "baseline_regional",
    ]
    for season_type in ["heating", "cooling"]:
        if season_type == "heating":
            weights = heating_weights
        else:
            weights = cooling_weights

        for filter_ in filters:
            stats_by_climate_zone = {
                cz: stats_dict.get("{}_{}_{}".format(cz, filter_, season_type))
                for cz in climate_zones
            }

            keys = ["percent_savings_{}".format(method) for method in methods]

            national_weightings = _compute_national_weightings(
                stats_by_climate_zone, keys, weights
            )

            bounds = _compute_national_weighting_lower_and_upper_bounds(
                stats_by_climate_zone, keys, weights
            )
            national_weightings.update(bounds)

            national_weightings.update(
                {"label": "national_weighted_mean_{}_{}".format(season_type, filter_)}
            )

            national_weighting_stats.append(national_weightings)

    stats = national_weighting_stats + stats
    return stats


def summary_statistics_to_csv(stats, filepath, product_id):
    """Write metric statistics to CSV file.

    Parameters
    ----------
    stats : list of dict
        List of outputs from thermostat.stats.compute_summary_statistics()
    filepath : str
        Filepath at which to save the suppary statistics
    product_id : str
        A combination of the connected thermostat service plus one or more
        connected thermostat device models that comprises the data set.

    Returns
    -------
    df : pandas.DataFrame
        A pandas dataframe containing the output data.

    """

    if stats is None:
        warnings.warn("No summary statistics to export.")
        return None

    drop_columns = [
        "national_weighted_mean_heating_no_filter",
        "national_weighted_mean_heating_tau_cvrmse_savings_p01_filter",
        "national_weighted_mean_cooling_no_filter",
        "national_weighted_mean_cooling_tau_cvrmse_savings_p01_filter",
        "national_weighted_mean_heating_tau_filter",
        "national_weighted_mean_heating_tau_cvrmse_filter",
        "national_weighted_mean_cooling_tau_filter",
        "national_weighted_mean_cooling_tau_cvrmse_filter",
    ]
    columns = [
        "label",
        "product_id",
        "sw_version",
    ]

    columns.extend(
        [
            "n_thermostat_core_day_sets_total",
            "n_thermostat_core_day_sets_kept",
            "n_thermostat_core_day_sets_discarded",
        ]
    )
    for column_name in REAL_OR_INTEGER_VALUED_COLUMNS_ALL:
        columns.append("{}_n".format(column_name))
        columns.append("{}_upper_bound_95_perc_conf".format(column_name))
        columns.append("{}_mean".format(column_name))
        columns.append("{}_lower_bound_95_perc_conf".format(column_name))
        columns.append("{}_sem".format(column_name))
        for quantile in QUANTILE:
            columns.append("{}_q{}".format(column_name, quantile))

    # add product_id
    for row in stats:
        row["product_id"] = product_id

    # transpose for readability.
    stats_dataframe = (
        pd.DataFrame(stats, columns=columns).set_index("label").transpose()
    )
    # Remove certification columns that are no longer in the stats file
    stats_dataframe.drop(drop_columns, axis=1, errors="ignore", inplace=True)
    stats_dataframe.to_csv(filepath)
    return stats_dataframe
